from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.forbidden import ForbiddenAndConjunction, ForbiddenEqualsClause
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    UniformFloatHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool, check_none


class LibLinear_Preprocessor(AutoSklearnPreprocessingAlgorithm):
    # Liblinear is not deterministic as it uses a RNG inside
    def __init__(
        self,
        penalty,
        loss,
        dual,
        tol,
        C,
        multi_class,
        fit_intercept,
        intercept_scaling,
        class_weight=None,
        random_state=None,
    ):
        self.penalty = penalty
        self.loss = loss
        self.dual = dual
        self.tol = tol
        self.C = C
        self.multi_class = multi_class
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.class_weight = class_weight
        self.random_state = random_state
        self.preprocessor = None

    def fit(self, X, Y):
        import sklearn.svm
        from sklearn.feature_selection import SelectFromModel

        self.C = float(self.C)
        self.tol = float(self.tol)
        self.dual = check_for_bool(self.dual)
        self.fit_intercept = check_for_bool(self.fit_intercept)
        self.intercept_scaling = float(self.intercept_scaling)

        if check_none(self.class_weight):
            self.class_weight = None

        estimator = sklearn.svm.LinearSVC(
            penalty=self.penalty,
            loss=self.loss,
            dual=self.dual,
            tol=self.tol,
            C=self.C,
            class_weight=self.class_weight,
            fit_intercept=self.fit_intercept,
            intercept_scaling=self.intercept_scaling,
            multi_class=self.multi_class,
            random_state=self.random_state,
        )

        estimator.fit(X, Y)
        self.preprocessor = SelectFromModel(
            estimator=estimator, threshold="mean", prefit=True
        )

        return self

    def transform(self, X):
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "LinearSVC Preprocessor",
            "name": "Liblinear Support Vector Classification Preprocessing",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "input": (SPARSE, DENSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()

        penalty = Constant("penalty", "l1")
        loss = CategoricalHyperparameter(
            "loss", ["hinge", "squared_hinge"], default_value="squared_hinge"
        )
        dual = Constant("dual", "False")
        # This is set ad-hoc
        tol = UniformFloatHyperparameter(
            "tol", 1e-5, 1e-1, default_value=1e-4, log=True
        )
        C = UniformFloatHyperparameter("C", 0.03125, 32768, log=True, default_value=1.0)
        multi_class = Constant("multi_class", "ovr")
        # These are set ad-hoc
        fit_intercept = Constant("fit_intercept", "True")
        intercept_scaling = Constant("intercept_scaling", 1)

        cs.add_hyperparameters(
            [penalty, loss, dual, tol, C, multi_class, fit_intercept, intercept_scaling]
        )

        penalty_and_loss = ForbiddenAndConjunction(
            ForbiddenEqualsClause(penalty, "l1"), ForbiddenEqualsClause(loss, "hinge")
        )
        cs.add_forbidden_clause(penalty_and_loss)
        return cs



import sklearn.ensemble

from autosklearn.pipeline.components.regression.extra_trees import ExtraTreesRegressor

from .test_base import BaseRegressionComponentTest


class ExtraTreesComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston"] = 0.8539264243687228
    res["boston_n_calls"] = 9
    res["default_boston_iterative"] = res["default_boston"]
    res["default_boston_sparse"] = 0.411211701806908
    res["default_boston_iterative_sparse"] = res["default_boston_sparse"]
    res["default_diabetes"] = 0.3885150255877827
    res["diabetes_n_calls"] = 9
    res["default_diabetes_iterative"] = res["default_diabetes"]
    res["default_diabetes_sparse"] = 0.2422804139169642
    res["default_diabetes_iterative_sparse"] = res["default_diabetes_sparse"]

    sk_mod = sklearn.ensemble.ExtraTreesRegressor
    module = ExtraTreesRegressor
    step_hyperparameter = {
        "name": "n_estimators",
        "value": module.get_max_iter(),
    }



from typing import Optional, Union

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from numpy.random import RandomState

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class RandomKitchenSinks(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        gamma: float,
        n_components: int,
        random_state: Optional[Union[int, RandomState]] = None,
    ) -> None:
        """
        Parameters
        ----------
        gamma: float
            Parameter of the rbf kernel to be approximated exp(-gamma * x^2)

        n_components: int
            Number of components (output dimensionality) used to approximate the kernel

        random_state: Optional[int | RandomState]
            The random state to pass to the underlying estimator
        """
        self.gamma = gamma
        self.n_components = n_components
        self.random_state = random_state

    def fit(self, X, Y=None):
        import sklearn.kernel_approximation

        self.n_components = int(self.n_components)
        self.gamma = float(self.gamma)

        self.preprocessor = sklearn.kernel_approximation.RBFSampler(
            gamma=self.gamma,
            n_components=self.n_components,
            random_state=self.random_state,
        )
        self.preprocessor.fit(X)
        return self

    def transform(self, X):
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "KitchenSink",
            "name": "Random Kitchen Sinks",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (SPARSE, DENSE, UNSIGNED_DATA),
            "output": (INPUT, UNSIGNED_DATA),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        gamma = UniformFloatHyperparameter(
            "gamma", 3.0517578125e-05, 8, default_value=1.0, log=True
        )
        n_components = UniformIntegerHyperparameter(
            "n_components", 50, 10000, default_value=100, log=True
        )
        cs = ConfigurationSpace()
        cs.add_hyperparameters([gamma, n_components])
        return cs



__author__ = "feurerm"

import copy

import numpy as np
import sklearn.datasets
import sklearn.metrics

from autosklearn.pipeline.classification import SimpleClassificationPipeline
from autosklearn.pipeline.components.classification.adaboost import AdaboostClassifier
from autosklearn.pipeline.components.classification.decision_tree import DecisionTree
from autosklearn.pipeline.components.classification.extra_trees import (
    ExtraTreesClassifier,
)
from autosklearn.pipeline.components.classification.gradient_boosting import (
    GradientBoostingClassifier,
)
from autosklearn.pipeline.components.classification.liblinear_svc import LibLinear_SVC
from autosklearn.pipeline.components.classification.libsvm_svc import LibSVM_SVC
from autosklearn.pipeline.components.classification.passive_aggressive import (
    PassiveAggressive,
)
from autosklearn.pipeline.components.classification.random_forest import RandomForest
from autosklearn.pipeline.components.classification.sgd import SGD
from autosklearn.pipeline.components.data_preprocessing.balancing.balancing import (
    Balancing,
)
from autosklearn.pipeline.components.feature_preprocessing.extra_trees_preproc_for_classification import (  # noqa: E501
    ExtraTreesPreprocessorClassification,
)
from autosklearn.pipeline.components.feature_preprocessing.liblinear_svc_preprocessor import (  # noqa: E501
    LibLinear_Preprocessor,
)

import unittest


class BalancingComponentTest(unittest.TestCase):
    def test_balancing_get_weights_treed_single_label(self):
        Y = np.array([0] * 80 + [1] * 20)
        balancing = Balancing(strategy="weighting")
        init_params, fit_params = balancing.get_weights(Y, "adaboost", None, None, None)
        self.assertAlmostEqual(
            np.mean(fit_params["classifier:sample_weight"]),
            1,
        )
        np.testing.assert_allclose(
            fit_params["classifier:sample_weight"],
            np.array([0.625] * 80 + [2.5] * 20),
        )

    def test_balancing_get_weights_treed_multilabel(self):
        Y = np.array(
            [[0, 0, 0]] * 100
            + [[1, 0, 0]] * 100
            + [[0, 1, 0]] * 100
            + [[1, 1, 0]] * 100
            + [[0, 0, 1]] * 100
            + [[1, 0, 1]] * 10
        )
        balancing = Balancing(strategy="weighting")
        init_params, fit_params = balancing.get_weights(Y, "adaboost", None, None, None)
        print(fit_params["classifier:sample_weight"])
        self.assertAlmostEqual(
            np.mean(fit_params["classifier:sample_weight"]),
            1,
        )
        np.testing.assert_allclose(
            fit_params["classifier:sample_weight"],
            np.array([0.85] * 500 + [8.5] * 10),
        )

    def test_balancing_get_weights_svm_sgd(self):
        Y = np.array([0] * 80 + [1] * 20)
        balancing = Balancing(strategy="weighting")
        init_params, fit_params = balancing.get_weights(
            Y, "libsvm_svc", None, None, None
        )
        self.assertEqual(
            ("classifier:class_weight", "balanced"), list(init_params.items())[0]
        )
        init_params, fit_params = balancing.get_weights(
            Y, None, "liblinear_svc_preprocessor", None, None
        )
        self.assertEqual(
            ("feature_preprocessor:class_weight", "balanced"),
            list(init_params.items())[0],
        )

    def test_weighting_effect(self):
        data = sklearn.datasets.make_classification(
            n_samples=200,
            n_features=10,
            n_redundant=2,
            n_informative=2,
            n_repeated=2,
            n_clusters_per_class=2,
            weights=[0.8, 0.2],
            random_state=1,
        )

        for name, clf, acc_no_weighting, acc_weighting, places in [
            ("adaboost", AdaboostClassifier, 0.810, 0.735, 3),
            ("decision_tree", DecisionTree, 0.780, 0.643, 3),
            ("extra_trees", ExtraTreesClassifier, 0.78, 0.8, 3),
            ("random_forest", RandomForest, 0.75, 0.789, 3),
            ("libsvm_svc", LibSVM_SVC, 0.769, 0.72, 3),
            ("liblinear_svc", LibLinear_SVC, 0.762, 0.735, 3),
            ("passive_aggressive", PassiveAggressive, 0.16, 0.222, 3),
            ("sgd", SGD, 0.818, 0.567, 2),
            ("gradient_boosting", GradientBoostingClassifier, 0.666, 0.682, 2),
        ]:
            for strategy, acc in [
                ("none", acc_no_weighting),
                ("weighting", acc_weighting),
            ]:
                # Fit
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                model_args = {
                    "random_state": 1,
                    "include": {
                        "classifier": [name],
                        "feature_preprocessor": ["no_preprocessing"],
                    },
                }

                classifier = SimpleClassificationPipeline(**model_args)
                cs = classifier.get_hyperparameter_search_space()
                default = cs.get_default_configuration()
                default._values["balancing:strategy"] = strategy

                classifier = SimpleClassificationPipeline(config=default, **model_args)
                classifier.fit(X_train, Y_train)

                predictions1 = classifier.predict(X_test)
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions1, Y_test),
                    acc,
                    places=places,
                    msg=(name, strategy),
                )

                # fit_transformer and fit_estimator
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                classifier = SimpleClassificationPipeline(config=default, **model_args)
                Xt, fit_params = classifier.fit_transformer(X_train, Y_train)
                classifier.fit_estimator(Xt, Y_train, **fit_params)

                predictions2 = classifier.predict(X_test)
                np.testing.assert_allclose(
                    predictions1,
                    predictions2,
                    err_msg=f"name = {name}, strategy = {strategy}",
                )
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions2, Y_test),
                    acc,
                    places=places,
                    msg=(name, strategy),
                )

        for name, pre, acc_no_weighting, acc_weighting in [
            (
                "extra_trees_preproc_for_classification",
                ExtraTreesPreprocessorClassification,
                0.810,
                0.590,
            ),
            ("liblinear_svc_preprocessor", LibLinear_Preprocessor, 0.837, 0.562),
        ]:
            for strategy, acc in [
                ("none", acc_no_weighting),
                ("weighting", acc_weighting),
            ]:
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                include = {"classifier": ["sgd"], "feature_preprocessor": [name]}

                classifier = SimpleClassificationPipeline(
                    random_state=1, include=include
                )
                cs = classifier.get_hyperparameter_search_space()
                default = cs.get_default_configuration()
                default._values["balancing:strategy"] = strategy
                classifier.set_hyperparameters(default)
                predictor = classifier.fit(X_train, Y_train)
                predictions = predictor.predict(X_test)
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions, Y_test),
                    acc,
                    places=3,
                    msg=(name, strategy),
                )

                # fit_transformer and fit_estimator
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                default._values["balancing:strategy"] = strategy
                classifier = SimpleClassificationPipeline(
                    config=default, random_state=1, include=include
                )
                Xt, fit_params = classifier.fit_transformer(X_train, Y_train)
                classifier.fit_estimator(Xt, Y_train, **fit_params)
                predictions = classifier.predict(X_test)
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions, Y_test), acc, places=3
                )



import time

from autosklearn.util.stopwatch import StopWatch


def test_stopwatch_overhead() -> None:
    wall_start = time.time()
    cpu_start = time.process_time()

    watch = StopWatch()
    for i in range(1, 1000):
        watch.start("task_%d" % i)
        watch.stop("task_%d" % i)

    cpu_end = time.process_time()
    wall_end = time.time()

    wall_duration = wall_end - wall_start
    cpu_duration = cpu_end - cpu_start

    cpu_overhead = cpu_duration - watch.total_cpu()
    wall_overhead = wall_duration - watch.total_wall()

    assert cpu_overhead < 1
    assert wall_overhead < 1
    assert watch.total_cpu() < 2 * watch.total_wall()


def test_contextmanager() -> None:
    watch = StopWatch()

    with watch.time("task"):
        assert watch["task"].started()

    assert "task" in watch
    assert watch["task"].finished()

    assert watch["task"].cpu_duration is not None
    assert watch["task"].wall_duration is not None



import sklearn.discriminant_analysis

from autosklearn.pipeline.components.classification.qda import QDA

from .test_base import BaseClassificationComponentTest


class QDAComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 1.0
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 0.56124476634783993
    res["default_iris_sparse"] = -1
    res["default_digits"] = 0.18882817243472982
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.89071038251366119
    res["default_digits_multilabel"] = 0.17011293429111091
    res["default_digits_multilabel_places"] = 1
    res["default_digits_multilabel_proba"] = 0.99999999999999989

    sk_mod = sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis
    module = QDA



# -*- encoding: utf-8 -*-
import typing

import glob
import os
import re

import numpy as np

scores_dict = {
    "train_single": ["single_best_train_score", "single_best_optimization_score"],
    "test_single": ["single_best_test_score"],
    "train_ensamble": ["ensemble_optimization_score"],
    "test_ensamble": ["ensemble_test_score"],
}


def print_debug_information(automl):
    # In case it is called with estimator,
    # Get the automl object
    if hasattr(automl, "automl_"):
        automl = automl.automl_

    # Log file path
    log_file = glob.glob(
        os.path.join(automl._backend.temporary_directory, "AutoML*.log")
    )[0]

    include_messages = ["INFO", "DEBUG", "WARN", "CRITICAL", "ERROR", "FATAL"]

    # There is a lot of content in the log files. Only
    # parsing the main message and ignore the metalearning
    # messages
    try:
        with open(log_file) as logfile:
            content = logfile.readlines()

        # Get the messages to debug easier!
        content = [
            line
            for line in content
            if any(msg in line for msg in include_messages)
            and "metalearning" not in line
        ]

    except Exception as e:
        return str(e)

    # Also add the run history if any
    if hasattr(automl, "runhistory_") and hasattr(automl.runhistory_, "data"):
        for k, v in automl.runhistory_.data.items():
            content += ["{}->{}".format(k, v)]
    else:
        content += ["No RunHistory"]

    # Also add the ensemble history if any
    if len(automl.ensemble_performance_history) > 0:
        content += [str(h) for h in automl.ensemble_performance_history]
    else:
        content += ["No Ensemble History"]

    return os.linesep.join(content)


def count_succeses(cv_results):
    return np.sum(
        [
            status in ["Success", "Success (but do not advance to higher budget)"]
            for status in cv_results["status"]
        ]
    )


def includes_all_scores(scores) -> bool:
    all_scores = (
        scores_dict["train_single"]
        + scores_dict["test_single"]
        + scores_dict["train_ensamble"]
        + scores_dict["test_ensamble"]
        + ["Timestamp"]
    )
    return set(scores) == set(all_scores)


def include_single_scores(scores) -> bool:
    all_scores = (
        scores_dict["train_single"] + scores_dict["test_single"] + ["Timestamp"]
    )
    return set(scores) == set(all_scores)


def includes_train_scores(scores) -> bool:
    all_scores = (
        scores_dict["train_single"] + scores_dict["train_ensamble"] + ["Timestamp"]
    )
    return set(scores) == set(all_scores)


def performance_over_time_is_plausible(poT) -> bool:
    if len(poT) < 1:
        return False
    if len(poT.drop(columns=["Timestamp"]).dropna()) < 1:
        return False
    if not poT["Timestamp"].is_monotonic:
        return False
    return True


class AutoMLLogParser(object):
    def __init__(self, logfile: str):
        self.logfile = logfile
        self.lines = self.parse_logfile()

    def parse_logfile(self) -> typing.List[str]:
        # We care about the [debug/info/...] messages
        # At the time of writing, the message format was:
        # [DEBUG] [2020-11-30 11:54:05,072:EnsembleBuilder] Restricting your
        # function to 3072 mb memory.
        #
        # [DEBUG] [2020-11-30 11:53:55,062:pynisher] Redirecting
        # output of the function to files.
        assert os.path.exists(self.logfile), "{} not found".format(self.logfile)

        with open(self.logfile) as fh:
            content = [line.strip() for line in fh if re.search(r"[\w+]", line)]
        return content

    def count_ensembler_iterations(self) -> int:
        iterations = []

        # One thing is to call phynisher, the other is to actually execute the funciton
        iterations_from_inside_ensemble_builder = []
        for line in self.lines:

            # Pynisher call
            # we have to count the start msg from pynisher
            # and the return msg
            # We expect the start msg to be something like:
            # [DEBUG] [2020-11-26 19:22:42,160:EnsembleBuilder] \
            # Function called with argument: (61....
            # [DEBUG] [2020-11-30 11:53:47,069:EnsembleBuilder] \
            # Function called with argument: (28.246965646743774, 1, False), {}
            match = re.search(
                r"EnsembleBuilder]\s+Function called with argument:\s+\(\d+\.\d+, (\d+), \w+",  # noqa: E501
                line,
            )
            if match:
                iterations.append(int(match.group(1)))

            # Ensemble Builder actual call
            # Here we expect the msg:
            # [DEBUG] [2020-11-30 11:53:14,877:EnsembleBuilder] Starting iteration 0,
            # time left: 61.266255
            # [DEBUG] [2020-11-27 20:27:28,044:EnsembleBuilder] Starting iteration 2,
            # time left: 10.603252
            match = re.search(r"EnsembleBuilder]\s+Starting iteration (\d+)", line)
            if match:
                iterations_from_inside_ensemble_builder.append(int(match.group(1)))

            # The ensemble builder might not be called if there is no time.
            # Here we expect the msg:
            # [DEBUG] [2020-11-27 20:27:28,044:EnsembleBuilder] \
            #   Not starting iteration 2, as time left: 1.59324
            match = re.search(r"EnsembleBuilder]\s+Not starting iteration (\d+)", line)
            if match:
                iterations_from_inside_ensemble_builder.append(int(match.group(1)))

        assert iterations == iterations_from_inside_ensemble_builder, "{} ! {}".format(
            iterations, iterations_from_inside_ensemble_builder
        )

        return iterations

    def count_ensembler_success_pynisher_calls(self) -> int:

        # We expect the return msg to be something like:
        # [DEBUG] [2020-11-30 11:53:47,911:EnsembleBuilder] return value:
        # (([{'Timestamp': Timestamp('2020-11-30 11:53:47.910727'),
        # 'ensemble_optimization_score': 0.9787234042553191}], 50, None, None, None), 0)
        # [DEBUG] [2020-11-30 11:54:05,984:EnsembleBuilder] return value:
        # (([{'Timestamp': Timestamp('2020-11- 30 11:54:05.983837'),
        # 'ensemble_optimization_score': 0.9787234042553191}], 50, None, None, None), 0)
        return_msgs = len(
            [
                line
                for line in self.lines
                if re.search(r"EnsembleBuilder]\s+return value:.*Timestamp", line)
            ]
        )

        return return_msgs

    def count_tae_pynisher_calls(self) -> int:
        # We expect the return msg to be something like:
        """
        [DEBUG] [2020-12-16 11:57:08,987:Client-pynisher] Function called with argument: (),
            {'queue': <multiprocessing.queues.Queue object at 0x7f9e3cfaae20>, 'config': 1
        [DEBUG] [2020-12-16 11:57:10,537:Client-pynisher] Function called with argument: (),
            {'queue': <multiprocessing.queues.Queue object at 0x7f16f5d95c40>, 'config': Configuration:
        """  # noqa: E501
        # Only the parenthesis below need to be escaped, ] and { do not.
        call_msgs = len(
            [
                line
                for line in self.lines
                if re.search(
                    r"pynisher]\s+Function called with argument: \(\), {", line
                )
            ]
        )
        return call_msgs

    def count_tae_pynisher_returns(self) -> int:
        # We expect the return msg to be something like:
        # [DEBUG] [2020-11-30 11:53:11,264:pynisher] return value: (None, 0)
        # [DEBUG] [2020-11-30 11:53:13,768:pynisher] return value: (None, 0)
        return_msgs = len(
            [
                line
                for line in self.lines
                if re.search(r"pynisher]\s+return value:\s+", line)
            ]
        )
        # When the pynisher pipe is prematurely closed, we also expect:
        # Your function call closed the pipe prematurely
        # -> Subprocess probably got an uncatchable signal
        # We expect the return msg to be something like:
        # OR
        # Something else went wrong, sorry.
        premature_msgs = len(
            [
                line
                for line in self.lines
                if re.search(
                    r"pynisher]\s+Your function call closed the pipe prematurely", line
                )
            ]
        )
        failure_msgs = len(
            [
                line
                for line in self.lines
                if re.search(r"pynisher]\s+Something else went wrong, sorry.", line)
            ]
        )
        return return_msgs + premature_msgs + failure_msgs

    def get_automl_setting_from_log(self, dataset_name: str, setting: str) -> str:
        for line in self.lines:
            # We expect messages of the form
            """
            [DEBUG] [2020-11-30 11:53:10,457:AutoML(5):breast_cancer]   ensemble_size: 50
            [DEBUG] [2020-11-30 11:53:10,457:AutoML(5):breast_cancer]   ensemble_nbest: 50
            """  # noqa: E501
            match = re.search(f"{dataset_name}]\\s*{setting}\\s*:\\s*(\\w+)", line)
            if match:
                return match.group(1)

        return None



from typing import Any, Dict, List, Optional, Sequence, Tuple, TypeVar, Union, cast

import copy
import json
import logging
import multiprocessing
import warnings

import numpy as np
import pandas
import pandas as pd
import scipy.sparse
from ConfigSpace import Configuration
from sklearn.base import BaseEstimator
from sklearn.model_selection import (
    BaseCrossValidator,
    KFold,
    PredefinedSplit,
    ShuffleSplit,
    StratifiedKFold,
    StratifiedShuffleSplit,
    train_test_split,
)
from sklearn.model_selection._split import BaseShuffleSplit, _RepeatedSplits
from smac.tae import StatusType, TAEAbortException

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.constants import (
    CLASSIFICATION_TASKS,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION_TASKS,
)
from autosklearn.data.abstract_data_manager import AbstractDataManager
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES, SUPPORTED_TARGET_TYPES
from autosklearn.evaluation.abstract_evaluator import (
    TYPE_ADDITIONAL_INFO,
    AbstractEvaluator,
    _fit_and_suppress_warnings,
)
from autosklearn.evaluation.splitter import (
    CustomStratifiedKFold,
    CustomStratifiedShuffleSplit,
)
from autosklearn.metrics import Scorer
from autosklearn.pipeline.base import PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import (
    IterativeComponent,
    ThirdPartyComponents,
)
from autosklearn.util.logging_ import PicklableClientLogger

__all__ = [
    "TrainEvaluator",
    "eval_holdout",
    "eval_iterative_holdout",
    "eval_cv",
    "eval_partial_cv",
    "eval_partial_cv_iterative",
]


def _get_y_array(y: SUPPORTED_TARGET_TYPES, task_type: int) -> SUPPORTED_TARGET_TYPES:
    if task_type in CLASSIFICATION_TASKS and task_type != MULTILABEL_CLASSIFICATION:
        return y.ravel()
    else:
        return y


T = TypeVar("T", SUPPORTED_FEAT_TYPES, SUPPORTED_TARGET_TYPES)


def select(data: T, indices: np.ndarray) -> T:
    """Select into some data by indices"""
    return data.iloc[indices] if hasattr(data, "iloc") else data[indices]


def subsample_indices(
    train_indices: List[int],
    subsample: Optional[float],
    task_type: int,
    Y_train: SUPPORTED_TARGET_TYPES,
) -> List[int]:

    if not isinstance(subsample, float):
        raise ValueError(
            "Subsample must be of type float, but is of type %s" % type(subsample)
        )
    elif subsample > 1:
        raise ValueError("Subsample must not be larger than 1, but is %f" % subsample)

    if subsample is not None and 0 <= subsample < 1:
        # Only subsample if there are more indices given to this method than
        # required to subsample because otherwise scikit-learn will complain

        if task_type in CLASSIFICATION_TASKS and task_type != MULTILABEL_CLASSIFICATION:
            stratify: Optional[SUPPORTED_TARGET_TYPES] = select(Y_train, train_indices)
        else:
            stratify = None

        indices = np.arange(len(train_indices))
        cv_indices_train, _ = train_test_split(
            indices,
            stratify=stratify,
            train_size=subsample,
            random_state=1,
            shuffle=True,
        )
        train_indices = train_indices[cv_indices_train]
        return train_indices

    return train_indices


def _fit_with_budget(
    X_train: SUPPORTED_FEAT_TYPES,
    Y_train: SUPPORTED_TARGET_TYPES,
    budget: float,
    budget_type: Optional[str],
    logger: Union[logging.Logger, PicklableClientLogger],
    model: BaseEstimator,
    train_indices: List[int],
    task_type: int,
) -> None:
    if budget_type == "iterations" or (
        budget_type == "mixed" and model.estimator_supports_iterative_fit()
    ):
        X = select(X_train, train_indices)
        y = select(Y_train, train_indices)

        if model.estimator_supports_iterative_fit():
            budget_factor = model.get_max_iter()
            Xt, fit_params = model.fit_transformer(X, y)

            n_iter = int(np.ceil(budget / 100 * budget_factor))
            model.iterative_fit(Xt, y, n_iter=n_iter, refit=True, **fit_params)
        else:
            _fit_and_suppress_warnings(logger, model, X, y)

    elif budget_type == "subsample" or (
        budget_type == "mixed" and not model.estimator_supports_iterative_fit()
    ):
        subsample = budget / 100
        train_indices_subset = subsample_indices(
            train_indices,
            subsample,
            task_type,
            Y_train,
        )
        X = select(X_train, train_indices_subset)
        y = select(Y_train, train_indices_subset)
        _fit_and_suppress_warnings(logger, model, X, y)

    else:
        raise ValueError(budget_type)


def concat_data(
    data: List[Any], num_cv_folds: int
) -> Union[np.ndarray, pd.DataFrame, scipy.sparse.csr_matrix]:
    if isinstance(data[0], np.ndarray):
        return np.concatenate(
            [data[i] for i in range(num_cv_folds) if data[i] is not None]
        )
    elif isinstance(data[0], scipy.sparse.spmatrix):
        return scipy.sparse.vstack(
            [data[i] for i in range(num_cv_folds) if data[i] is not None]
        )
    elif isinstance(data[0], pd.DataFrame):
        return pd.concat(
            [data[i] for i in range(num_cv_folds) if data[i] is not None],
            axis=0,
        )
    else:
        raise ValueError(f"Unknown datatype {type(data[0])}")


class TrainEvaluator(AbstractEvaluator):
    def __init__(
        self,
        backend: Backend,
        queue: multiprocessing.Queue,
        metrics: Sequence[Scorer],
        additional_components: Dict[str, ThirdPartyComponents],
        port: Optional[int],
        configuration: Optional[Union[int, Configuration]] = None,
        scoring_functions: Optional[List[Scorer]] = None,
        seed: int = 1,
        output_y_hat_optimization: bool = True,
        resampling_strategy: Optional[
            Union[str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit]
        ] = None,
        resampling_strategy_args: Optional[
            Dict[str, Optional[Union[float, int, str]]]
        ] = None,
        num_run: Optional[int] = None,
        budget: Optional[float] = None,
        budget_type: Optional[str] = None,
        keep_models: bool = False,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        disable_file_output: bool = False,
        init_params: Optional[Dict[str, Any]] = None,
    ):

        super().__init__(
            backend=backend,
            queue=queue,
            port=port,
            configuration=configuration,
            metrics=metrics,
            additional_components=additional_components,
            scoring_functions=scoring_functions,
            seed=seed,
            output_y_hat_optimization=output_y_hat_optimization,
            num_run=num_run,
            include=include,
            exclude=exclude,
            disable_file_output=disable_file_output,
            init_params=init_params,
            budget=budget,
            budget_type=budget_type,
        )

        self.feat_type = self.backend.load_datamanager().feat_type
        self.resampling_strategy = resampling_strategy
        if resampling_strategy_args is None:
            self.resampling_strategy_args = {}
        else:
            self.resampling_strategy_args = resampling_strategy_args
        self.splitter = self.get_splitter(self.datamanager)
        self.num_cv_folds = self.splitter.get_n_splits(
            groups=self.resampling_strategy_args.get("groups")
        )
        self.X_train = self.datamanager.data["X_train"]
        self.Y_train = self.datamanager.data["Y_train"]
        self.X_targets = [None] * self.num_cv_folds
        self.Y_targets = [None] * self.num_cv_folds
        self.Y_train_targets = np.ones(self.Y_train.shape) * np.NaN
        self.models = [None] * self.num_cv_folds
        self.indices: List[Optional[Tuple[List[int], List[int]]]] = [
            None
        ] * self.num_cv_folds

        # Necessary for full CV. Makes full CV not write predictions if only
        # a subset of folds is evaluated but time is up. Complicated, because
        #  code must also work for partial CV, where we want exactly the
        # opposite.
        self.partial = True
        self.keep_models = keep_models

    def fit_predict_and_loss(self, iterative: bool = False) -> None:
        """Fit, predict and compute the loss for cross-validation and
        holdout (both iterative and non-iterative)
        """
        # Define beforehand for mypy
        additional_run_info: Optional[TYPE_ADDITIONAL_INFO] = None

        if iterative:
            if self.num_cv_folds == 1:

                for train_split, test_split in self.splitter.split(
                    self.X_train,
                    self.Y_train,
                    groups=self.resampling_strategy_args.get("groups"),
                ):
                    self.X_optimization = select(self.X_train, test_split)
                    self.Y_optimization = select(self.Y_train, test_split)
                    self.Y_actual_train = select(self.Y_train, train_split)
                    self._partial_fit_and_predict_iterative(
                        0,
                        train_indices=train_split,
                        test_indices=test_split,
                        add_model_to_self=True,
                    )
            else:

                # Test if the model allows for an iterative fit, if not,
                # call this method again without the iterative argument
                model = self._get_model(self.feat_type)
                if not model.estimator_supports_iterative_fit():
                    self.fit_predict_and_loss(iterative=False)
                    return

                self.partial = False

                converged = [False] * self.num_cv_folds

                Y_train_pred = [None] * self.num_cv_folds
                Y_optimization_pred = [None] * self.num_cv_folds
                Y_test_pred = [None] * self.num_cv_folds
                train_splits = [None] * self.num_cv_folds

                self.models = [
                    self._get_model(self.feat_type) for i in range(self.num_cv_folds)
                ]
                iterations = [1] * self.num_cv_folds
                total_n_iterations = [0] * self.num_cv_folds
                # model.estimator_supports_iterative_fit -> true
                # After the if above, we know estimator support iterative fit
                model_max_iter = [
                    cast(IterativeComponent, model).get_max_iter()
                    for model in self.models
                ]

                if self.budget_type in ["iterations", "mixed"] and self.budget is None:
                    raise ValueError(
                        f"When budget type is {self.budget_type} the budget "
                        "can not be None"
                    )

                if (
                    self.budget_type in ["iterations", "mixed"]
                    and cast(float, self.budget) > 0
                ):
                    max_n_iter_budget = int(
                        np.ceil(cast(float, self.budget) / 100 * model_max_iter[0])
                    )
                    max_iter = min(model_max_iter[0], max_n_iter_budget)
                else:
                    max_iter = model_max_iter[0]

                models_current_iters = [0] * self.num_cv_folds

                Xt_array = [None] * self.num_cv_folds
                fit_params_array = [
                    {}
                ] * self.num_cv_folds  # type: List[Dict[str, Any]]

                y = _get_y_array(self.Y_train, self.task_type)

                # stores train loss(es) of each fold.
                train_losses = [dict()] * self.num_cv_folds
                # used as weights when averaging train losses.
                train_fold_weights = [np.NaN] * self.num_cv_folds
                # stores opt (validation) loss of each fold.
                opt_losses = [np.NaN] * self.num_cv_folds
                # weights for opt_losses.
                opt_fold_weights = [np.NaN] * self.num_cv_folds

                while not all(converged):

                    splitter = self.get_splitter(self.datamanager)

                    for i, (train_indices, test_indices) in enumerate(
                        splitter.split(
                            self.X_train,
                            y,
                            groups=self.resampling_strategy_args.get("groups"),
                        )
                    ):
                        if converged[i]:
                            continue

                        model = self.models[i]

                        if iterations[i] == 1:
                            self.Y_train_targets[train_indices] = select(
                                self.Y_train, train_indices
                            )
                            self.X_targets[i] = select(self.X_train, test_indices)
                            self.Y_targets[i] = select(self.Y_train, test_indices)

                            # Note: Be careful moving these into variables, caused a
                            # headache when trying to debug why things were breaking
                            Xt, fit_params = model.fit_transformer(
                                select(self.X_train, train_indices),
                                select(self.Y_train, train_indices),
                            )
                            Xt_array[i] = Xt
                            fit_params_array[i] = fit_params

                        n_iter = int(2 ** iterations[i] / 2) if iterations[i] > 1 else 2
                        total_n_iterations[i] = total_n_iterations[i] + n_iter

                        model.iterative_fit(
                            Xt_array[i],
                            select(self.Y_train, train_indices),
                            n_iter=n_iter,
                            **fit_params_array[i],
                        )

                        (train_pred, opt_pred, test_pred) = self._predict(
                            model,
                            train_indices=train_indices,
                            test_indices=test_indices,
                        )

                        Y_train_pred[i] = train_pred
                        Y_optimization_pred[i] = opt_pred
                        Y_test_pred[i] = test_pred
                        train_splits[i] = train_indices

                        train_loss = self._loss(
                            select(self.Y_train, train_indices),
                            train_pred,
                            X_data=Xt_array[i],
                        )
                        train_losses[i] = train_loss
                        # Number of training data points for this fold.
                        # Used for weighting the average.
                        train_fold_weights[i] = len(train_indices)

                        # Compute validation loss of this fold and store it.
                        optimization_loss = self._loss(
                            self.Y_targets[i], opt_pred, X_data=self.X_targets[i]
                        )
                        opt_losses[i] = optimization_loss
                        # number of optimization data points for this fold.
                        # Used for weighting the average.
                        opt_fold_weights[i] = len(test_indices)

                        models_current_iters[i] = model.get_current_iter()

                        if (
                            model.configuration_fully_fitted()
                            or models_current_iters[i] >= max_iter
                        ):
                            converged[i] = True

                        iterations[i] = iterations[i] + 1

                    # Compute weights of each fold based on the number of samples
                    # in each fold.
                    train_fold_weights_percentage = [
                        w / sum(train_fold_weights) for w in train_fold_weights
                    ]
                    opt_fold_weights_percentage = [
                        w / sum(opt_fold_weights) for w in opt_fold_weights
                    ]

                    train_loss = {
                        metric.name: np.average(
                            [
                                train_losses[i][str(metric)]
                                for i in range(self.num_cv_folds)
                            ],
                            weights=train_fold_weights_percentage,
                        )
                        for metric in self.metrics
                    }

                    # if all_scoring_function is true, return a dict of opt_loss.
                    # Otherwise, return a scalar.
                    opt_loss = {}
                    for metric in opt_losses[0].keys():
                        opt_loss[metric] = np.average(
                            [opt_losses[i][metric] for i in range(self.num_cv_folds)],
                            weights=opt_fold_weights_percentage,
                        )

                    X_targets = self.X_targets
                    Y_targets = self.Y_targets
                    Y_train_targets = self.Y_train_targets

                    Y_optimization_pred_concat = concat_data(
                        Y_optimization_pred, num_cv_folds=self.num_cv_folds
                    )
                    X_targets = concat_data(X_targets, num_cv_folds=self.num_cv_folds)
                    Y_targets = concat_data(Y_targets, num_cv_folds=self.num_cv_folds)

                    if self.X_test is not None:
                        Y_test_preds = np.array(
                            [
                                Y_test_pred[i]
                                for i in range(self.num_cv_folds)
                                if Y_test_pred[i] is not None
                            ]
                        )
                        # Average the predictions of several models
                        if len(Y_test_preds.shape) == 3:
                            Y_test_preds = np.nanmean(Y_test_preds, axis=0)
                    else:
                        Y_test_preds = None

                    self.X_optimization = X_targets
                    self.Y_optimization = Y_targets
                    self.Y_actual_train = Y_train_targets

                    self.model = self._get_model(self.feat_type)
                    status = StatusType.DONOTADVANCE
                    if any(
                        [
                            model_current_iter == max_iter
                            for model_current_iter in models_current_iters
                        ]
                    ):
                        status = StatusType.SUCCESS
                    self.finish_up(
                        loss=opt_loss,
                        train_loss=train_loss,
                        opt_pred=Y_optimization_pred_concat,
                        test_pred=Y_test_preds,
                        additional_run_info=additional_run_info,
                        file_output=True,
                        final_call=all(converged),
                        status=status,
                    )

        else:

            self.partial = False

            Y_train_pred = [None] * self.num_cv_folds
            Y_optimization_pred = [None] * self.num_cv_folds
            Y_test_pred = [None] * self.num_cv_folds
            train_splits = [None] * self.num_cv_folds

            y = _get_y_array(self.Y_train, self.task_type)

            train_losses = []  # stores train loss of each fold.
            train_fold_weights = []  # used as weights when averaging train losses.
            opt_losses = []  # stores opt (validation) loss of each fold.
            opt_fold_weights = []  # weights for opt_losses.

            # TODO: mention that no additional run info is possible in this
            # case! -> maybe remove full CV from the train evaluator anyway and
            # make the user implement this!
            for i, (train_split, test_split) in enumerate(
                self.splitter.split(
                    self.X_train, y, groups=self.resampling_strategy_args.get("groups")
                )
            ):

                # TODO add check that split is actually an integer array,
                # not a boolean array (to allow indexed assignement of
                # training data later).

                if self.budget_type is None:
                    (
                        train_pred,
                        opt_pred,
                        test_pred,
                        additional_run_info,
                    ) = self._partial_fit_and_predict_standard(
                        i,
                        train_indices=train_split,
                        test_indices=test_split,
                        add_model_to_self=self.num_cv_folds == 1,
                    )
                else:
                    (
                        train_pred,
                        opt_pred,
                        test_pred,
                        additional_run_info,
                    ) = self._partial_fit_and_predict_budget(
                        i,
                        train_indices=train_split,
                        test_indices=test_split,
                        add_model_to_self=self.num_cv_folds == 1,
                    )

                if (
                    additional_run_info is not None
                    and len(additional_run_info) > 0
                    and i > 0
                ):
                    raise TAEAbortException(
                        'Found additional run info "%s" in fold %d, '
                        "but cannot handle additional run info if fold >= 1."
                        % (additional_run_info, i)
                    )

                Y_train_pred[i] = train_pred
                Y_optimization_pred[i] = opt_pred
                Y_test_pred[i] = test_pred
                train_splits[i] = train_split

                X = select(self.X_train, train_split)
                y = select(self.Y_train_targets, train_split)

                train_loss = self._loss(y, train_pred, X_data=X)
                train_losses.append(train_loss)
                # number of training data points for this fold. Used for weighting
                # the average.
                train_fold_weights.append(len(train_split))

                # Compute validation loss of this fold and store it.
                optimization_loss = self._loss(
                    self.Y_targets[i],
                    opt_pred,
                    X_data=self.X_targets[i],
                )
                opt_losses.append(optimization_loss)
                # number of optimization data points for this fold. Used for weighting
                # the average.
                opt_fold_weights.append(len(test_split))

            # Compute weights of each fold based on the number of samples in each
            # fold.
            train_fold_weights = [
                w / sum(train_fold_weights) for w in train_fold_weights
            ]
            opt_fold_weights = [w / sum(opt_fold_weights) for w in opt_fold_weights]

            train_loss = {
                metric.name: np.average(
                    [train_losses[i][str(metric)] for i in range(self.num_cv_folds)],
                    weights=train_fold_weights,
                )
                for metric in self.metrics
            }

            # if all_scoring_function is true, return a dict of opt_loss. Otherwise,
            # return a scalar.
            opt_loss = {}
            for metric_name in list(opt_losses[0].keys()) + [
                metric.name for metric in self.metrics
            ]:
                opt_loss[metric_name] = np.average(
                    [opt_losses[i][metric_name] for i in range(self.num_cv_folds)],
                    weights=opt_fold_weights,
                )

            X_targets = self.X_targets
            Y_targets = self.Y_targets
            Y_train_targets = self.Y_train_targets

            Y_optimization_pred = concat_data(
                Y_optimization_pred, num_cv_folds=self.num_cv_folds
            )
            X_targets = concat_data(X_targets, num_cv_folds=self.num_cv_folds)
            Y_targets = concat_data(Y_targets, num_cv_folds=self.num_cv_folds)

            if self.X_test is not None:
                Y_test_pred = np.array(
                    [
                        Y_test_pred[i]
                        for i in range(self.num_cv_folds)
                        if Y_test_pred[i] is not None
                    ]
                )
                # Average the predictions of several models
                if len(np.shape(Y_test_pred)) == 3:
                    Y_test_pred = np.nanmean(Y_test_pred, axis=0)

            self.X_optimization = X_targets
            self.Y_optimization = Y_targets
            self.Y_actual_train = Y_train_targets

            if self.num_cv_folds > 1:
                self.model = self._get_model(self.feat_type)
                # Bad style, but necessary for unit testing that self.model is
                # actually a new model
                self._added_empty_model = True
                # TODO check if there might be reasons for do-not-advance here!
                status = StatusType.SUCCESS
            elif (
                self.budget_type == "iterations"
                or self.budget_type == "mixed"
                and self.model.estimator_supports_iterative_fit()
            ):
                budget_factor = self.model.get_max_iter()
                # We check for budget being None in initialization
                n_iter = int(np.ceil(cast(float, self.budget) / 100 * budget_factor))
                model_current_iter = self.model.get_current_iter()
                if model_current_iter < n_iter:
                    status = StatusType.DONOTADVANCE
                else:
                    status = StatusType.SUCCESS
            else:
                if self.model.estimator_supports_iterative_fit():
                    model_max_iter = self.model.get_max_iter()
                    model_current_iter = self.model.get_current_iter()
                    if model_current_iter < model_max_iter:
                        status = StatusType.DONOTADVANCE
                    else:
                        status = StatusType.SUCCESS
                else:
                    status = StatusType.SUCCESS

            self.finish_up(
                loss=opt_loss,
                train_loss=train_loss,
                opt_pred=Y_optimization_pred,
                test_pred=Y_test_pred if self.X_test is not None else None,
                additional_run_info=additional_run_info,
                file_output=True,
                final_call=True,
                status=status,
            )

    def partial_fit_predict_and_loss(self, fold: int, iterative: bool = False) -> None:
        """Fit, predict and get loss for eval_partial_cv (iterative and normal)"""
        if fold > self.num_cv_folds:
            raise ValueError(
                "Cannot evaluate a fold %d which is higher than "
                "the number of folds %d." % (fold, self.num_cv_folds)
            )
        if self.budget_type is not None:
            raise NotImplementedError()

        y = _get_y_array(self.Y_train, self.task_type)
        for i, (train_split, test_split) in enumerate(
            self.splitter.split(
                self.X_train, y, groups=self.resampling_strategy_args.get("groups")
            )
        ):
            if i != fold:
                continue
            else:
                break

        if self.num_cv_folds > 1:
            self.X_optimization = select(self.X_train, test_split)
            self.Y_optimization = select(self.Y_train, test_split)
            self.Y_actual_train = select(self.Y_train, train_split)

        if iterative:
            self._partial_fit_and_predict_iterative(
                fold,
                train_indices=train_split,
                test_indices=test_split,
                add_model_to_self=True,
            )
        elif self.budget_type is not None:
            raise NotImplementedError()
        else:
            (
                train_pred,
                opt_pred,
                test_pred,
                additional_run_info,
            ) = self._partial_fit_and_predict_standard(
                fold,
                train_indices=train_split,
                test_indices=test_split,
                add_model_to_self=True,
            )

            # This is my best guess at what the X_data for these should be
            X_train = select(self.X_train, train_split)  # From above (only cv?)
            X_fold = self.X_targets[fold]
            train_loss = self._loss(self.Y_actual_train, train_pred, X_data=X_train)
            loss = self._loss(self.Y_targets[fold], opt_pred, X_data=X_fold)

            if self.model.estimator_supports_iterative_fit():
                model_max_iter = self.model.get_max_iter()
                model_current_iter = self.model.get_current_iter()
                if model_current_iter < model_max_iter:
                    status = StatusType.DONOTADVANCE
                else:
                    status = StatusType.SUCCESS
            else:
                status = StatusType.SUCCESS

            self.finish_up(
                loss=loss,
                train_loss=train_loss,
                opt_pred=opt_pred,
                test_pred=test_pred,
                file_output=False,
                final_call=True,
                additional_run_info=None,
                status=status,
            )

    def _partial_fit_and_predict_iterative(
        self,
        fold: int,
        train_indices: List[int],
        test_indices: List[int],
        add_model_to_self: bool,
    ) -> None:
        model = self._get_model(self.feat_type)

        self.indices[fold] = (train_indices, test_indices)

        # Do only output the files in the case of iterative holdout,
        # In case of iterative partial cv, no file output is needed
        # because ensembles cannot be built
        file_output = True if self.num_cv_folds == 1 else False

        if model.estimator_supports_iterative_fit():
            X = select(self.X_train, train_indices)
            y = select(self.Y_train, train_indices)

            X_test = select(self.X_train, test_indices)
            y_test = select(self.Y_train, test_indices)

            Xt, fit_params = model.fit_transformer(X, y)

            self.Y_train_targets[train_indices] = y

            iteration = 1
            total_n_iteration = 0
            model_max_iter = model.get_max_iter()

            if self.budget is not None and self.budget > 0:
                max_n_iter_budget = int(np.ceil(self.budget / 100 * model_max_iter))
                max_iter = min(model_max_iter, max_n_iter_budget)
            else:
                max_iter = model_max_iter
            model_current_iter = 0

            while (
                not model.configuration_fully_fitted() and model_current_iter < max_iter
            ):
                n_iter = int(2**iteration / 2) if iteration > 1 else 2
                total_n_iteration += n_iter
                model.iterative_fit(Xt, y, n_iter=n_iter, **fit_params)

                (Y_train_pred, Y_optimization_pred, Y_test_pred,) = self._predict(
                    model,
                    train_indices=train_indices,
                    test_indices=test_indices,
                )

                if add_model_to_self:
                    self.model = model

                train_loss = self._loss(y, Y_train_pred, X_data=X)
                loss = self._loss(y_test, Y_optimization_pred, X_data=X_test)

                additional_run_info = model.get_additional_run_info()

                model_current_iter = model.get_current_iter()
                if model_current_iter < max_iter:
                    status = StatusType.DONOTADVANCE
                else:
                    status = StatusType.SUCCESS

                if model.configuration_fully_fitted() or model_current_iter >= max_iter:
                    final_call = True
                else:
                    final_call = False

                self.finish_up(
                    loss=loss,
                    train_loss=train_loss,
                    opt_pred=Y_optimization_pred,
                    test_pred=Y_test_pred,
                    additional_run_info=additional_run_info,
                    file_output=file_output,
                    final_call=final_call,
                    status=status,
                )
                iteration += 1

            return
        else:

            (
                Y_train_pred,
                Y_optimization_pred,
                Y_test_pred,
                additional_run_info,
            ) = self._partial_fit_and_predict_standard(
                fold, train_indices, test_indices, add_model_to_self
            )

            X = select(self.X_train, train_indices)
            y = select(self.Y_train, train_indices)
            train_loss = self._loss(y, Y_train_pred, X_data=X)

            X_test = select(self.X_train, test_indices)
            y_test = select(self.Y_train, test_indices)
            loss = self._loss(y_test, Y_optimization_pred, X_data=X_test)

            if self.model.estimator_supports_iterative_fit():
                model_max_iter = self.model.get_max_iter()
                model_current_iter = self.model.get_current_iter()
                if model_current_iter < model_max_iter:
                    status = StatusType.DONOTADVANCE
                else:
                    status = StatusType.SUCCESS
            else:
                status = StatusType.SUCCESS
            self.finish_up(
                loss=loss,
                train_loss=train_loss,
                opt_pred=Y_optimization_pred,
                test_pred=Y_test_pred,
                additional_run_info=additional_run_info,
                file_output=file_output,
                final_call=True,
                status=status,
            )
            return

    def _partial_fit_and_predict_standard(
        self,
        fold: int,
        train_indices: List[int],
        test_indices: List[int],
        add_model_to_self: bool = False,
    ) -> Tuple[
        PIPELINE_DATA_DTYPE,  # train_pred
        PIPELINE_DATA_DTYPE,  # opt_pred
        PIPELINE_DATA_DTYPE,  # test_pred
        TYPE_ADDITIONAL_INFO,
    ]:
        model = self._get_model(self.feat_type)

        self.indices[fold] = (train_indices, test_indices)

        X = select(self.X_train, train_indices)
        y = select(self.Y_train, train_indices)
        _fit_and_suppress_warnings(self.logger, model, X, y)

        if add_model_to_self:
            self.model = model
        else:
            self.models[fold] = model

        self.X_targets[fold] = select(self.X_train, test_indices)
        self.Y_targets[fold] = select(self.Y_train, test_indices)
        self.Y_train_targets[train_indices] = select(self.Y_train, train_indices)

        train_pred, opt_pred, test_pred = self._predict(
            model=model,
            train_indices=train_indices,
            test_indices=test_indices,
        )
        additional_run_info = model.get_additional_run_info()
        return (
            train_pred,
            opt_pred,
            test_pred,
            additional_run_info,
        )

    def _partial_fit_and_predict_budget(
        self,
        fold: int,
        train_indices: List[int],
        test_indices: List[int],
        add_model_to_self: bool = False,
    ) -> Tuple[
        PIPELINE_DATA_DTYPE,  # train_pred
        PIPELINE_DATA_DTYPE,  # opt_pred
        PIPELINE_DATA_DTYPE,  # test_pred
        TYPE_ADDITIONAL_INFO,
    ]:

        # This function is only called in the event budget is not None
        # Add this statement for mypy
        assert self.budget is not None

        model = self._get_model(self.feat_type)
        self.indices[fold] = (train_indices, test_indices)
        self.X_targets[fold] = select(self.X_train, test_indices)
        self.Y_targets[fold] = select(self.Y_train, test_indices)
        self.Y_train_targets[train_indices] = select(self.Y_train, train_indices)

        _fit_with_budget(
            X_train=self.X_train,
            Y_train=self.Y_train,
            budget=self.budget,
            budget_type=self.budget_type,
            logger=self.logger,
            model=model,
            train_indices=train_indices,
            task_type=self.task_type,
        )

        train_pred, opt_pred, test_pred = self._predict(
            model,
            train_indices=train_indices,
            test_indices=test_indices,
        )

        if add_model_to_self:
            self.model = model
        else:
            self.models[fold] = model

        additional_run_info = model.get_additional_run_info()
        return (
            train_pred,
            opt_pred,
            test_pred,
            additional_run_info,
        )

    def _predict(
        self, model: BaseEstimator, test_indices: List[int], train_indices: List[int]
    ) -> Tuple[PIPELINE_DATA_DTYPE, PIPELINE_DATA_DTYPE, PIPELINE_DATA_DTYPE]:
        y_train = select(self.Y_train, train_indices)
        X_train = select(self.X_train, train_indices)
        X_test = select(self.X_train, test_indices)

        # The y_train here does not correspond to the X, there to ensure output shape
        # will be correct in case labels were missing from some split
        train_pred = self.predict_function(X_train, model, self.task_type, y_train)
        opt_pred = self.predict_function(X_test, model, self.task_type, y_train)

        # This is the test data the user can pass in
        if self.X_test is not None:
            # See comment above about y_train
            X_user = self.X_test.copy()
            test_pred = self.predict_function(X_user, model, self.task_type, y_train)
        else:
            test_pred = None

        return train_pred, opt_pred, test_pred

    def get_splitter(
        self, D: AbstractDataManager
    ) -> Union[BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit]:

        if self.resampling_strategy_args is None:
            self.resampling_strategy_args = {}

        if self.resampling_strategy is not None and not isinstance(
            self.resampling_strategy, str
        ):
            if "groups" not in self.resampling_strategy_args:
                self.resampling_strategy_args["groups"] = None

            if isinstance(
                self.resampling_strategy,
                (BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit),
            ):
                self.check_splitter_resampling_strategy(
                    X=D.data["X_train"],
                    y=D.data["Y_train"],
                    groups=self.resampling_strategy_args.get("groups"),
                    task=D.info["task"],
                    resampling_strategy=self.resampling_strategy,
                )
                return self.resampling_strategy

            # If it got to this point, we are dealing with a non-supported
            # re-sampling strategy
            raise ValueError(
                "Unsupported resampling strategy {}/{} provided".format(
                    self.resampling_strategy,
                    type(self.resampling_strategy),
                )
            )

        y = D.data["Y_train"]
        shuffle = self.resampling_strategy_args.get("shuffle", True)
        train_size = 0.67
        if self.resampling_strategy_args:
            train_size_from_user = self.resampling_strategy_args.get("train_size")
            if train_size_from_user is not None:
                train_size = float(train_size_from_user)
        test_size = float("%.4f" % (1 - train_size))

        if (
            D.info["task"] in CLASSIFICATION_TASKS
            and D.info["task"] != MULTILABEL_CLASSIFICATION
        ):

            y = y.ravel()
            if self.resampling_strategy in ["holdout", "holdout-iterative-fit"]:

                if shuffle:
                    try:
                        cv = StratifiedShuffleSplit(
                            n_splits=1,
                            test_size=test_size,
                            random_state=1,
                        )
                        test_cv = copy.deepcopy(cv)
                        next(test_cv.split(y, y))
                    except ValueError as e:
                        if "The least populated class in y has only" in e.args[0]:
                            cv = CustomStratifiedShuffleSplit(
                                n_splits=1,
                                test_size=test_size,
                                random_state=1,
                            )
                        else:
                            raise e
                else:
                    tmp_train_size = int(np.floor(train_size * y.shape[0]))
                    test_fold = np.zeros(y.shape[0])
                    test_fold[:tmp_train_size] = -1
                    cv = PredefinedSplit(test_fold=test_fold)
                    cv.n_splits = 1  # As sklearn is inconsistent here
            elif self.resampling_strategy in [
                "cv",
                "cv-iterative-fit",
                "partial-cv",
                "partial-cv-iterative-fit",
            ]:
                if shuffle:
                    try:
                        with warnings.catch_warnings():
                            warnings.simplefilter("error")
                            cv = StratifiedKFold(
                                n_splits=self.resampling_strategy_args["folds"],
                                shuffle=shuffle,
                                random_state=1,
                            )
                            test_cv = copy.deepcopy(cv)
                            next(test_cv.split(y, y))
                    except UserWarning as e:
                        print(e)
                        if "The least populated class in y has only" in e.args[0]:
                            cv = CustomStratifiedKFold(
                                n_splits=self.resampling_strategy_args["folds"],
                                shuffle=shuffle,
                                random_state=1,
                            )
                        else:
                            raise e
                else:
                    cv = KFold(
                        n_splits=self.resampling_strategy_args["folds"], shuffle=shuffle
                    )
            else:
                raise ValueError(self.resampling_strategy)
        else:
            if self.resampling_strategy in ["holdout", "holdout-iterative-fit"]:
                # TODO shuffle not taken into account for this
                if shuffle:
                    cv = ShuffleSplit(n_splits=1, test_size=test_size, random_state=1)
                else:
                    tmp_train_size = int(np.floor(train_size * y.shape[0]))
                    test_fold = np.zeros(y.shape[0])
                    test_fold[:tmp_train_size] = -1
                    cv = PredefinedSplit(test_fold=test_fold)
                    cv.n_splits = 1  # As sklearn is inconsistent here
            elif self.resampling_strategy in [
                "cv",
                "partial-cv",
                "partial-cv-iterative-fit",
            ]:
                random_state = 1 if shuffle else None
                cv = KFold(
                    n_splits=self.resampling_strategy_args["folds"],
                    shuffle=shuffle,
                    random_state=random_state,
                )
            else:
                raise ValueError(self.resampling_strategy)
        return cv

    @classmethod
    def check_splitter_resampling_strategy(
        cls,
        X: PIPELINE_DATA_DTYPE,
        y: np.ndarray,
        task: int,
        groups: Any,
        resampling_strategy: Union[
            BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
        ],
    ) -> None:
        if (
            task in CLASSIFICATION_TASKS
            and task != MULTILABEL_CLASSIFICATION
            or (task in REGRESSION_TASKS and task != MULTIOUTPUT_REGRESSION)
        ):
            y = y.ravel()

        try:
            resampling_strategy.get_n_splits(X=X, y=y, groups=groups)
            next(resampling_strategy.split(X=X, y=y, groups=groups))
        except Exception as e:
            raise ValueError(
                "Unsupported resampling strategy "
                "{}/{} cause exception: {}".format(
                    resampling_strategy,
                    groups,
                    str(e),
                )
            )


# create closure for evaluating an algorithm
def eval_holdout(
    queue: multiprocessing.Queue,
    config: Union[int, Configuration],
    backend: Backend,
    resampling_strategy: Union[
        str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
    ],
    resampling_strategy_args: Dict[str, Optional[Union[float, int, str]]],
    metrics: Sequence[Scorer],
    seed: int,
    num_run: int,
    instance: str,
    scoring_functions: Optional[List[Scorer]],
    output_y_hat_optimization: bool,
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    disable_file_output: bool,
    port: Optional[int],
    additional_components: Dict[str, ThirdPartyComponents],
    init_params: Optional[Dict[str, Any]] = None,
    budget: Optional[float] = 100.0,
    budget_type: Optional[str] = None,
    iterative: bool = False,
) -> None:
    evaluator = TrainEvaluator(
        backend=backend,
        port=port,
        queue=queue,
        resampling_strategy=resampling_strategy,
        resampling_strategy_args=resampling_strategy_args,
        metrics=metrics,
        configuration=config,
        seed=seed,
        num_run=num_run,
        scoring_functions=scoring_functions,
        output_y_hat_optimization=output_y_hat_optimization,
        include=include,
        exclude=exclude,
        disable_file_output=disable_file_output,
        additional_components=additional_components,
        init_params=init_params,
        budget=budget,
        budget_type=budget_type,
    )
    evaluator.fit_predict_and_loss(iterative=iterative)


def eval_iterative_holdout(
    queue: multiprocessing.Queue,
    config: Union[int, Configuration],
    backend: Backend,
    resampling_strategy: Union[
        str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
    ],
    resampling_strategy_args: Dict[str, Optional[Union[float, int, str]]],
    metrics: Sequence[Scorer],
    seed: int,
    num_run: int,
    instance: str,
    scoring_functions: Optional[List[Scorer]],
    output_y_hat_optimization: bool,
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    disable_file_output: bool,
    port: Optional[int],
    additional_components: Dict[str, ThirdPartyComponents],
    init_params: Optional[Dict[str, Any]] = None,
    budget: Optional[float] = 100.0,
    budget_type: Optional[str] = None,
) -> None:
    return eval_holdout(
        queue=queue,
        port=port,
        config=config,
        backend=backend,
        metrics=metrics,
        resampling_strategy=resampling_strategy,
        resampling_strategy_args=resampling_strategy_args,
        seed=seed,
        num_run=num_run,
        scoring_functions=scoring_functions,
        output_y_hat_optimization=output_y_hat_optimization,
        include=include,
        exclude=exclude,
        instance=instance,
        disable_file_output=disable_file_output,
        iterative=True,
        additional_components=additional_components,
        init_params=init_params,
        budget=budget,
        budget_type=budget_type,
    )


def eval_partial_cv(
    queue: multiprocessing.Queue,
    config: Union[int, Configuration],
    backend: Backend,
    resampling_strategy: Union[
        str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
    ],
    resampling_strategy_args: Dict[str, Optional[Union[float, int, str]]],
    metrics: Sequence[Scorer],
    seed: int,
    num_run: int,
    instance: str,
    scoring_functions: Optional[List[Scorer]],
    output_y_hat_optimization: bool,
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    disable_file_output: bool,
    port: Optional[int],
    additional_components: Dict[str, ThirdPartyComponents],
    init_params: Optional[Dict[str, Any]] = None,
    budget: Optional[float] = None,
    budget_type: Optional[str] = None,
    iterative: bool = False,
) -> None:
    if budget_type is not None:
        raise NotImplementedError()
    instance_dict: Dict[str, int] = json.loads(instance) if instance is not None else {}
    fold = instance_dict["fold"]

    evaluator = TrainEvaluator(
        backend=backend,
        port=port,
        queue=queue,
        metrics=metrics,
        configuration=config,
        resampling_strategy=resampling_strategy,
        resampling_strategy_args=resampling_strategy_args,
        seed=seed,
        num_run=num_run,
        scoring_functions=scoring_functions,
        output_y_hat_optimization=False,
        include=include,
        exclude=exclude,
        disable_file_output=disable_file_output,
        additional_components=additional_components,
        init_params=init_params,
        budget=budget,
        budget_type=budget_type,
    )

    evaluator.partial_fit_predict_and_loss(fold=fold, iterative=iterative)


def eval_partial_cv_iterative(
    queue: multiprocessing.Queue,
    config: Union[int, Configuration],
    backend: Backend,
    resampling_strategy: Union[
        str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
    ],
    resampling_strategy_args: Dict[str, Optional[Union[float, int, str]]],
    metrics: Sequence[Scorer],
    seed: int,
    num_run: int,
    instance: str,
    scoring_functions: Optional[List[Scorer]],
    output_y_hat_optimization: bool,
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    disable_file_output: bool,
    port: Optional[int],
    additional_components: Dict[str, ThirdPartyComponents],
    init_params: Optional[Dict[str, Any]] = None,
    budget: Optional[float] = None,
    budget_type: Optional[str] = None,
) -> None:
    if budget_type is not None:
        raise NotImplementedError()
    return eval_partial_cv(
        queue=queue,
        config=config,
        backend=backend,
        metrics=metrics,
        resampling_strategy=resampling_strategy,
        resampling_strategy_args=resampling_strategy_args,
        seed=seed,
        port=port,
        num_run=num_run,
        instance=instance,
        scoring_functions=scoring_functions,
        output_y_hat_optimization=output_y_hat_optimization,
        include=include,
        exclude=exclude,
        disable_file_output=disable_file_output,
        iterative=True,
        additional_components=additional_components,
        init_params=init_params,
    )


# create closure for evaluating an algorithm
def eval_cv(
    queue: multiprocessing.Queue,
    config: Union[int, Configuration],
    backend: Backend,
    resampling_strategy: Union[
        str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
    ],
    resampling_strategy_args: Dict[str, Optional[Union[float, int, str]]],
    metrics: Sequence[Scorer],
    seed: int,
    num_run: int,
    instance: str,
    scoring_functions: Optional[List[Scorer]],
    output_y_hat_optimization: bool,
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    disable_file_output: bool,
    port: Optional[int],
    additional_components: Dict[str, ThirdPartyComponents],
    init_params: Optional[Dict[str, Any]] = None,
    budget: Optional[float] = None,
    budget_type: Optional[str] = None,
    iterative: bool = False,
) -> None:
    evaluator = TrainEvaluator(
        backend=backend,
        port=port,
        queue=queue,
        metrics=metrics,
        configuration=config,
        seed=seed,
        num_run=num_run,
        resampling_strategy=resampling_strategy,
        resampling_strategy_args=resampling_strategy_args,
        scoring_functions=scoring_functions,
        output_y_hat_optimization=output_y_hat_optimization,
        include=include,
        exclude=exclude,
        disable_file_output=disable_file_output,
        additional_components=additional_components,
        init_params=init_params,
        budget=budget,
        budget_type=budget_type,
    )

    evaluator.fit_predict_and_loss(iterative=iterative)


def eval_iterative_cv(
    queue: multiprocessing.Queue,
    config: Union[int, Configuration],
    backend: Backend,
    resampling_strategy: Union[
        str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit
    ],
    resampling_strategy_args: Dict[str, Optional[Union[float, int, str]]],
    metrics: Sequence[Scorer],
    seed: int,
    num_run: int,
    instance: str,
    scoring_functions: Optional[List[Scorer]],
    output_y_hat_optimization: bool,
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    disable_file_output: bool,
    port: Optional[int],
    additional_components: Dict[str, ThirdPartyComponents],
    init_params: Optional[Dict[str, Any]] = None,
    budget: Optional[float] = None,
    budget_type: Optional[str] = None,
    iterative: bool = True,
) -> None:
    eval_cv(
        backend=backend,
        queue=queue,
        metrics=metrics,
        config=config,
        seed=seed,
        num_run=num_run,
        resampling_strategy=resampling_strategy,
        resampling_strategy_args=resampling_strategy_args,
        scoring_functions=scoring_functions,
        output_y_hat_optimization=output_y_hat_optimization,
        include=include,
        exclude=exclude,
        disable_file_output=disable_file_output,
        port=port,
        additional_components=additional_components,
        init_params=init_params,
        budget=budget,
        budget_type=budget_type,
        iterative=iterative,
        instance=instance,
    )



import json
import os
import random
import shutil
import socket
import subprocess
import tempfile

import arff
import numpy as np

from autosklearn.metrics import CLASSIFICATION_METRICS, REGRESSION_METRICS

import unittest

# Abstracted to here to make changing them easier
# seems to be quite fidely with github actions
# Recommended to set a minimum of 60
timeouts = {
    1: 60,  # create commands
    2: 180,  # generate metadata
    3: 60,  # get performance of configurations
    4: 90,  # Calculate metafeatures
    5: 60,  # Create aslib files
}


class TestMetadataGeneration(unittest.TestCase):
    def setUp(self):
        host = socket.gethostname()
        pid = os.getpid()
        rint = random.randint(0, 1000000)

        self.working_directory = os.path.join(
            tempfile.gettempdir(), f"autosklearn-unittest-tmp-dir-{host}-{pid}-{rint}"
        )

    def print_files(self):
        print("Existing files:")
        for dirpath, dirnames, filenames in os.walk(self.working_directory):
            print(dirpath, dirnames, filenames)

    def test_metadata_generation(self):
        regression_task_id = 360029
        regression_dataset_name = "SWD".lower()
        classification_task_id = 245
        classification_dataset_name = "breast-w".lower()

        current_directory = __file__
        scripts_directory = os.path.abspath(
            os.path.join(current_directory, "..", "..", "..", "scripts")
        )

        # 1. create working directory
        try:
            os.makedirs(self.working_directory)
        except Exception as e:
            print(e)

        # 2. should be done by the person running the unit tests!

        # 3. create configuration commands
        script_filename = os.path.join(scripts_directory, "01_create_commands.py")
        cmd = "python3 %s --working-directory %s --test" % (
            script_filename,
            self.working_directory,
        )

        return_value = subprocess.run(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=30,
        )
        self.assertEqual(return_value.returncode, 0, msg=f"{cmd}\n{str(return_value)}")

        # 4. run one of the commands to get some data
        commands_output_file = os.path.join(
            self.working_directory, "metadata_commands.txt"
        )
        self.assertTrue(os.path.exists(commands_output_file))

        with open(commands_output_file) as fh:
            cmds = fh.read().split("\n")
            # 6 regression, 7 classification (roc_auc + task 258 is illegal),
            # 1 empty line
            self.assertEqual(len(cmds), 18, msg="\n".join(cmds))

        for task_id, dataset_name, task_type, metric in (
            (
                classification_task_id,
                classification_dataset_name,
                "classification",
                "balanced_accuracy",
            ),
            (regression_task_id, regression_dataset_name, "regression", "r2"),
        ):
            cmd = None
            with open(commands_output_file) as fh:
                while True:
                    cmd = fh.readline()
                    if "task-id %d" % task_id in cmd and metric in cmd:
                        break
            if cmd is None:
                self.fail(
                    "Did not find a command for task_id %s and metric %s in %s"
                    % (task_id, metric, cmds)
                )

            self.assertIn("time-limit 86400", cmd)
            self.assertIn("per-run-time-limit 1800", cmd)
            cmd = cmd.replace("time-limit 86400", "time-limit 60").replace(
                "per-run-time-limit 1800", "per-run-time-limit 5"
            )
            # This tells the script to use the same memory limit for testing as
            # for training. In production, it would use twice as much!
            cmd = cmd.replace("-s 1", "-s 1 --unittest")
            print("COMMAND: %s" % cmd)

            return_value = subprocess.run(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=180,
            )
            print("STDOUT: %s" % repr(return_value.stdout), flush=True)
            print("STDERR: %s" % repr(return_value.stderr), flush=True)

            self.print_files()

            expected_output_directory = os.path.join(
                self.working_directory,
                "configuration",
                task_type,
                str(task_id),
                metric,
                "auto-sklearn-output",
            )
            self.assertTrue(
                os.path.exists(expected_output_directory), msg=expected_output_directory
            )
            smac_log = os.path.join(
                expected_output_directory, "AutoML(1):%s.log" % dataset_name
            )
            with open(smac_log) as fh:
                smac_output = fh.read()
            self.assertEqual(
                return_value.returncode,
                0,
                msg=f"{cmd}\n{str(return_value)}" + "\n" + smac_output,
            )
            expected_validation_output = os.path.join(
                expected_output_directory, "..", "validation_trajectory_1.json"
            )
            self.assertTrue(os.path.exists(expected_validation_output))
            trajectory = os.path.join(
                expected_output_directory, "smac3-output", "run_1", "trajectory.json"
            )

            with open(expected_validation_output) as fh_validation:
                with open(trajectory) as fh_trajectory:
                    traj = json.load(fh_trajectory)
                    valid_traj = json.load(fh_validation)
                    print("Validation trajectory:")
                    print(valid_traj)
                    self.assertGreater(len(traj), 2, msg=str(valid_traj))
                    self.assertEqual(len(traj), len(valid_traj), msg=str(valid_traj))
                    for entry in valid_traj:
                        if task_type == "classification":
                            for metric in CLASSIFICATION_METRICS:
                                # This is a multilabel metric
                                if metric in (
                                    "precision_samples",
                                    "recall_samples",
                                    "f1_samples",
                                ):
                                    continue
                                self.assertIn(metric, entry[-1])
                                self.assertIsInstance(entry[-1][metric], float)
                                self.assertTrue(
                                    np.isfinite(entry[-1][metric]),
                                    (metric, str(entry[-1][metric])),
                                )
                        else:
                            for metric in REGRESSION_METRICS:
                                self.assertIn(metric, entry[-1])
                                self.assertIsInstance(entry[-1][metric], float)
                                self.assertTrue(
                                    np.isfinite(entry[-1][metric]),
                                    (metric, str(entry[-1][metric])),
                                )

        # 5. Get the test performance of these configurations
        script_filename = os.path.join(scripts_directory, "02_retrieve_metadata.py")
        cmd = "python3 %s --working-directory %s " % (
            script_filename,
            self.working_directory,
        )
        print("COMMAND: %s" % cmd)

        return_value = subprocess.run(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=60,
        )
        print("STDOUT: %s" % repr(return_value.stdout), flush=True)
        print("STDERR: %s" % repr(return_value.stderr), flush=True)
        self.assertEqual(return_value.returncode, 0, msg=f"{cmd}\n{str(return_value)}")

        for file in [
            "algorithm_runs.arff",
            "configurations.csv",
            "description.results.txt",
        ]:
            for metric in ["accuracy", "balanced_accuracy", "log_loss", "roc_auc"]:
                path = os.path.join(
                    self.working_directory,
                    "configuration_results",
                    "%s_binary.classification_dense" % metric,
                    file,
                )
                self.assertTrue(os.path.exists(path), msg=path)

        for file in [
            "algorithm_runs.arff",
            "configurations.csv",
            "description.results.txt",
        ]:
            for metric in ["r2", "mean_squared_error"]:
                path = os.path.join(
                    self.working_directory,
                    "configuration_results",
                    "%s_regression_dense" % metric,
                    file,
                )
                self.assertTrue(os.path.exists(path), msg=path)

        # 6. Calculate metafeatures
        script_filename = os.path.join(
            scripts_directory, "03_calculate_metafeatures.py"
        )
        cmd = "python3 %s --working-directory %s --test-mode " % (
            script_filename,
            self.working_directory,
        )
        return_value = subprocess.run(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=180,
        )
        self.assertEqual(return_value.returncode, 0, msg=f"{cmd}\n{str(return_value)}")
        for task_type in ("classification", "regression"):
            for file in [
                "calculation_times.csv",
                "description.features.txt",
                "feature_costs.arff",
                "feature_runstatus.arff",
                "feature_values.arff",
            ]:
                self.assertTrue(
                    os.path.exists(
                        os.path.join(
                            self.working_directory, "metafeatures", task_type, file
                        )
                    )
                )

        with open(
            os.path.join(
                self.working_directory,
                "metafeatures",
                "regression",
                "feature_values.arff",
            )
        ) as fh:
            metafeatures_arff = fh.read().split("\n")
            contains_regression_id = False
            for line in metafeatures_arff:
                if line.startswith("fri_c4_500_25,"):
                    contains_regression_id = True
            self.assertTrue(contains_regression_id, msg=metafeatures_arff)

        with open(
            os.path.join(
                self.working_directory,
                "metafeatures",
                "classification",
                "feature_values.arff",
            )
        ) as fh:
            metafeatures_arff = fh.read().split("\n")
            contains_classification_id = False
            for line in metafeatures_arff:
                if line.startswith("anneal,"):
                    contains_classification_id = True
            self.assertTrue(contains_classification_id, msg=metafeatures_arff)

        # 7. Create aslib files
        script_filename = os.path.join(scripts_directory, "04_create_aslib_files.py")
        cmd = "python3 %s --working-directory %s " % (
            script_filename,
            self.working_directory,
        )

        return_value = subprocess.run(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=45,
        )
        self.assertEqual(return_value.returncode, 0, msg=f"{cmd}\n{str(return_value)}")

        for metric_, combination in (
            (metric, "%s_binary.classification_dense" % metric),
            (metric, "%s_regression_dense" % metric),
        ):

            if task_type not in combination:
                continue

            for file in [
                "algorithm_runs.arff",
                "configurations.csv",
                "description.txt",
                "feature_costs.arff",
                "feature_runstatus.arff",
                "feature_values.arff",
                "readme.txt",
            ]:
                expected_path = os.path.join(
                    self.working_directory,
                    "metadata",
                    combination,
                    file,
                )
                self.assertTrue(os.path.exists(expected_path), msg=expected_path)

            with open(
                os.path.join(
                    self.working_directory,
                    "metadata",
                    combination,
                    "algorithm_runs.arff",
                )
            ) as fh:
                algorithm_runs = arff.load(fh)
                self.assertEqual(
                    algorithm_runs["attributes"],
                    [
                        ("instance_id", "STRING"),
                        ("repetition", "NUMERIC"),
                        ("algorithm", "STRING"),
                        (metric_, "NUMERIC"),
                        (
                            "runstatus",
                            [
                                "ok",
                                "timeout",
                                "memout",
                                "not_applicable",
                                "crash",
                                "other",
                            ],
                        ),
                    ],
                )
                self.assertEqual(len(algorithm_runs["data"]), 1)
                self.assertEqual(len(algorithm_runs["data"][0]), 5)
                self.assertLess(algorithm_runs["data"][0][3], 0.9)
                self.assertEqual(algorithm_runs["data"][0][4], "ok")

    def tearDown(self):
        for i in range(5):
            try:
                shutil.rmtree(self.working_directory)
            except Exception:
                pass



import sklearn.metrics
import sklearn.svm

from autosklearn.pipeline.components.classification.libsvm_svc import LibSVM_SVC
from autosklearn.pipeline.util import _test_classifier_predict_proba, get_dataset

from .test_base import BaseClassificationComponentTest


class LibSVM_SVCComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.96
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 0.36298338197586716
    res["default_iris_sparse"] = 0.64
    res["default_digits"] = 0.096539162112932606
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.90103217972070426
    res["default_digits_multilabel"] = -1
    res["default_digits_multilabel_proba"] = -1

    sk_mod = sklearn.svm.SVC
    module = LibSVM_SVC

    def test_default_configuration_predict_proba_individual(self):
        # Leave this additional test here
        for i in range(2):
            predictions, targets = _test_classifier_predict_proba(
                LibSVM_SVC, sparse=True, dataset="digits", train_size_maximum=500
            )
            self.assertAlmostEqual(
                5.273502056835706, sklearn.metrics.log_loss(targets, predictions)
            )

        for i in range(2):
            predictions, targets = _test_classifier_predict_proba(
                LibSVM_SVC, sparse=True, dataset="iris"
            )
            self.assertAlmostEqual(
                0.8408320837510618, sklearn.metrics.log_loss(targets, predictions)
            )

        # 2 class
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(dataset="iris")
            remove_training_data = Y_train == 2
            remove_test_data = Y_test == 2
            X_train = X_train[~remove_training_data]
            Y_train = Y_train[~remove_training_data]
            X_test = X_test[~remove_test_data]
            Y_test = Y_test[~remove_test_data]
            ss = sklearn.preprocessing.StandardScaler()
            X_train = ss.fit_transform(X_train)
            configuration_space = LibSVM_SVC.get_hyperparameter_search_space()
            default = configuration_space.get_default_configuration()

            cls = LibSVM_SVC(
                random_state=1,
                **{
                    hp_name: default[hp_name]
                    for hp_name in default
                    if default[hp_name] is not None
                },
            )

            cls = cls.fit(X_train, Y_train)
            prediction = cls.predict_proba(X_test)
            self.assertAlmostEqual(
                sklearn.metrics.log_loss(Y_test, prediction),
                0.6927962762794081,
                places=4,
            )



import operator

import numpy as np


def pareto_front(values: np.ndarray, *, is_loss: bool = True) -> np.ndarray:
    """Calculate the pareto front

    source from: https://stackoverflow.com/a/40239615

    Note
    ----
    Works on the assumption that every value is either something to minimize or
    something to maximize, based on ``is_loss``.

    Parameters
    ----------
    values: np.ndarray [n_models, n_metrics]
        The value for each of the metrics

    is_loss: bool = True
        Whether the metrics are a loss or a score

    Returns
    -------
    np.ndarray
        A boolean mask where true indicates if the model on the pareto front
    """
    op = operator.lt if is_loss else operator.gt

    is_efficient = np.ones(values.shape[0], dtype=bool)
    for i, c in enumerate(values):
        if is_efficient[i]:
            # Keep any point with a lower cost
            is_efficient[is_efficient] = np.any(op(values[is_efficient], c), axis=1)

            # And keep self
            is_efficient[i] = True

    return is_efficient



from .abstract_ensemble import AbstractEnsemble, AbstractMultiObjectiveEnsemble
from .ensemble_selection import EnsembleSelection
from .multiobjective_dummy_ensemble import MultiObjectiveDummyEnsemble
from .singlebest_ensemble import (
    SingleBest,
    SingleBestFromRunhistory,
    SingleModelEnsemble,
)

__all__ = [
    "AbstractEnsemble",
    "AbstractMultiObjectiveEnsemble",
    "EnsembleSelection",
    "SingleBestFromRunhistory",
    "SingleBest",
    "SingleModelEnsemble",
    "MultiObjectiveDummyEnsemble",
]



import numpy as np

from autosklearn.pipeline.components.feature_preprocessing.pca import PCA
from autosklearn.pipeline.util import PreprocessingTestCase, _test_preprocessing


class PCAComponentTest(PreprocessingTestCase):
    def test_default_configuration(self):
        transformations = []
        for i in range(2):
            transformation, original = _test_preprocessing(PCA)
            self.assertEqual(transformation.shape, original.shape)
            self.assertFalse((transformation == original).all())
            transformations.append(transformation)
            if len(transformations) > 1:
                np.testing.assert_allclose(
                    transformations[-1], transformations[-2], rtol=1e-4
                )

    def test_preprocessing_dtype(self):
        super(PCAComponentTest, self)._test_preprocessing_dtype(PCA, test_sparse=False)



from typing import Optional, Union

import copy
from itertools import product

import numpy as np
from ConfigSpace.configuration_space import Configuration, ConfigurationSpace
from ConfigSpace.forbidden import ForbiddenAndConjunction, ForbiddenEqualsClause
from sklearn.base import ClassifierMixin

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import BasePipeline
from autosklearn.pipeline.components.classification import ClassifierChoice
from autosklearn.pipeline.components.data_preprocessing import DataPreprocessorChoice
from autosklearn.pipeline.components.data_preprocessing.balancing.balancing import (
    Balancing,
)
from autosklearn.pipeline.components.feature_preprocessing import (
    FeaturePreprocessorChoice,
)
from autosklearn.pipeline.constants import SPARSE


class SimpleClassificationPipeline(BasePipeline, ClassifierMixin):
    """This class implements the classification task.

    It implements a pipeline, which includes one preprocessing step and one
    classification algorithm. It can render a search space including all known
    classification and preprocessing algorithms.

    Contrary to the sklearn API it is not possible to enumerate the
    possible parameters in the __init__ function because we only know the
    available classifiers at runtime. For this reason the user must
    specifiy the parameters by passing an instance of
    ConfigSpace.configuration_space.Configuration.

    Parameters
    ----------
    config : ConfigSpace.configuration_space.Configuration
        The configuration to evaluate.

    random_state : Optional[int | RandomState]
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance
        used by `np.random`.

    Attributes
    ----------
    _estimator : The underlying scikit-learn classification model. This
        variable is assigned after a call to the
        :meth:`autosklearn.pipeline.classification.SimpleClassificationPipeline
        .fit` method.

    _preprocessor : The underlying scikit-learn preprocessing algorithm. This
        variable is only assigned if a preprocessor is specified and
        after a call to the
        :meth:`autosklearn.pipeline.classification.SimpleClassificationPipeline
        .fit` method.

    See also
    --------

    References
    ----------

    Examples
    --------

    """

    def __init__(
        self,
        config: Optional[Configuration] = None,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        steps=None,
        dataset_properties=None,
        include=None,
        exclude=None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        init_params=None,
    ):
        self._output_dtype = np.int32
        if dataset_properties is None:
            dataset_properties = dict()
        if "target_type" not in dataset_properties:
            dataset_properties["target_type"] = "classification"
        super().__init__(
            feat_type=feat_type,
            config=config,
            steps=steps,
            dataset_properties=dataset_properties,
            include=include,
            exclude=exclude,
            random_state=random_state,
            init_params=init_params,
        )

    def fit_transformer(self, X, y, fit_params=None):

        if fit_params is None:
            fit_params = {}

        if self.config["balancing:strategy"] == "weighting":
            balancing = Balancing(strategy="weighting")
            _init_params, _fit_params = balancing.get_weights(
                y,
                self.config["classifier:__choice__"],
                self.config["feature_preprocessor:__choice__"],
                {},
                {},
            )
            _init_params.update(self.init_params)
            self.set_hyperparameters(
                feat_type=self.feat_type,
                configuration=self.config,
                init_params=_init_params,
            )

            if _fit_params is not None:
                fit_params.update(_fit_params)

        X, fit_params = super().fit_transformer(X, y, fit_params=fit_params)

        return X, fit_params

    def predict_proba(self, X, batch_size=None):
        """predict_proba.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        batch_size: int or None, defaults to None
            batch_size controls whether the pipeline will be
            called on small chunks of the data. Useful when calling the
            predict method on the whole array X results in a MemoryError.

        Returns
        -------
        array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)
        """
        if batch_size is None:
            return super().predict_proba(X)

        else:
            if not isinstance(batch_size, int):
                raise ValueError(
                    "Argument 'batch_size' must be of type int, "
                    "but is '%s'" % type(batch_size)
                )
            if batch_size <= 0:
                raise ValueError(
                    "Argument 'batch_size' must be positive, " "but is %d" % batch_size
                )

            else:
                # Probe for the target array dimensions
                target = self.predict_proba(X[0:2].copy())

                y = np.zeros((X.shape[0], target.shape[1]), dtype=np.float32)

                for k in range(max(1, int(np.ceil(float(X.shape[0]) / batch_size)))):
                    batch_from = k * batch_size
                    batch_to = min([(k + 1) * batch_size, X.shape[0]])
                    pred_prob = self.predict_proba(
                        X[batch_from:batch_to], batch_size=None
                    )
                    y[batch_from:batch_to] = pred_prob.astype(np.float32)

                return y

    def _get_hyperparameter_search_space(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        include=None,
        exclude=None,
        dataset_properties=None,
    ):
        """Create the hyperparameter configuration space.

        Parameters
        ----------
        feat_type : dict, maps columns to there datatypes

        include : dict (optional, default=None)

        Returns
        -------
        cs : ConfigSpace.configuration_space.Configuration
            The configuration space describing the SimpleRegressionClassifier.
        """
        cs = ConfigurationSpace()

        if dataset_properties is None or not isinstance(dataset_properties, dict):
            dataset_properties = dict()
        if "target_type" not in dataset_properties:
            dataset_properties["target_type"] = "classification"
        if dataset_properties["target_type"] != "classification":
            dataset_properties["target_type"] = "classification"

        if "sparse" not in dataset_properties:
            # This dataset is probably dense
            dataset_properties["sparse"] = False

        cs = self._get_base_search_space(
            cs=cs,
            feat_type=feat_type,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=self.steps,
        )

        classifiers = cs.get_hyperparameter("classifier:__choice__").choices
        preprocessors = cs.get_hyperparameter("feature_preprocessor:__choice__").choices
        available_classifiers = self._final_estimator.get_available_components(
            dataset_properties
        )

        possible_default_classifier = copy.copy(list(available_classifiers.keys()))
        default = cs.get_hyperparameter("classifier:__choice__").default_value
        del possible_default_classifier[possible_default_classifier.index(default)]

        # A classifier which can handle sparse data after the densifier is
        # forbidden for memory issues
        for key in classifiers:
            if SPARSE in available_classifiers[key].get_properties()["input"]:
                if "densifier" in preprocessors:
                    while True:
                        try:
                            forb_cls = ForbiddenEqualsClause(
                                cs.get_hyperparameter("classifier:__choice__"), key
                            )
                            forb_fpp = ForbiddenEqualsClause(
                                cs.get_hyperparameter(
                                    "feature_preprocessor:__choice__"
                                ),
                                "densifier",
                            )
                            cs.add_forbidden_clause(
                                ForbiddenAndConjunction(forb_cls, forb_fpp)
                            )
                            # Success
                            break
                        except ValueError:
                            # Change the default and try again
                            try:
                                default = possible_default_classifier.pop()
                            except IndexError:
                                raise ValueError(
                                    "Cannot find a legal default configuration."
                                )
                            cs.get_hyperparameter(
                                "classifier:__choice__"
                            ).default_value = default

        # which would take too long
        # Combinations of non-linear models with feature learning:
        classifiers_ = [
            "adaboost",
            "decision_tree",
            "extra_trees",
            "gradient_boosting",
            "k_nearest_neighbors",
            "libsvm_svc",
            "mlp",
            "random_forest",
            "gaussian_nb",
        ]
        feature_learning = [
            "kernel_pca",
            "kitchen_sinks",
            "nystroem_sampler",
        ]

        for c, f in product(classifiers_, feature_learning):
            if c not in classifiers:
                continue
            if f not in preprocessors:
                continue
            while True:
                try:
                    cs.add_forbidden_clause(
                        ForbiddenAndConjunction(
                            ForbiddenEqualsClause(
                                cs.get_hyperparameter("classifier:__choice__"), c
                            ),
                            ForbiddenEqualsClause(
                                cs.get_hyperparameter(
                                    "feature_preprocessor:__choice__"
                                ),
                                f,
                            ),
                        )
                    )
                    break
                except KeyError:
                    break
                except ValueError:
                    # Change the default and try again
                    try:
                        default = possible_default_classifier.pop()
                    except IndexError:
                        raise ValueError("Cannot find a legal default configuration.")
                    cs.get_hyperparameter(
                        "classifier:__choice__"
                    ).default_value = default

        # Won't work
        # Multinomial NB etc don't use with features learning, pca etc
        classifiers_ = ["multinomial_nb"]
        preproc_with_negative_X = [
            "kitchen_sinks",
            "pca",
            "truncatedSVD",
            "fast_ica",
            "kernel_pca",
            "nystroem_sampler",
        ]

        for c, f in product(classifiers_, preproc_with_negative_X):
            if c not in classifiers:
                continue
            if f not in preprocessors:
                continue
            while True:
                try:
                    cs.add_forbidden_clause(
                        ForbiddenAndConjunction(
                            ForbiddenEqualsClause(
                                cs.get_hyperparameter(
                                    "feature_preprocessor:__choice__"
                                ),
                                f,
                            ),
                            ForbiddenEqualsClause(
                                cs.get_hyperparameter("classifier:__choice__"), c
                            ),
                        )
                    )
                    break
                except KeyError:
                    break
                except ValueError:
                    # Change the default and try again
                    try:
                        default = possible_default_classifier.pop()
                    except IndexError:
                        raise ValueError("Cannot find a legal default configuration.")
                    cs.get_hyperparameter(
                        "classifier:__choice__"
                    ).default_value = default

        self.configuration_space = cs
        self.dataset_properties = dataset_properties
        return cs

    def _get_pipeline_steps(
        self, dataset_properties, feat_type: Optional[FEAT_TYPE_TYPE] = None
    ):
        steps = []

        default_dataset_properties = {"target_type": "classification"}
        if dataset_properties is not None and isinstance(dataset_properties, dict):
            default_dataset_properties.update(dataset_properties)

        steps.extend(
            [
                [
                    "data_preprocessor",
                    DataPreprocessorChoice(
                        feat_type=feat_type,
                        dataset_properties=default_dataset_properties,
                        random_state=self.random_state,
                    ),
                ],
                ["balancing", Balancing(random_state=self.random_state)],
                [
                    "feature_preprocessor",
                    FeaturePreprocessorChoice(
                        feat_type=feat_type,
                        dataset_properties=default_dataset_properties,
                        random_state=self.random_state,
                    ),
                ],
                [
                    "classifier",
                    ClassifierChoice(
                        feat_type=feat_type,
                        dataset_properties=default_dataset_properties,
                        random_state=self.random_state,
                    ),
                ],
            ]
        )

        return steps

    def _get_estimator_hyperparameter_name(self):
        return "classifier"



import sklearn.ensemble

from autosklearn.pipeline.components.classification.extra_trees import (
    ExtraTreesClassifier,
)

from .test_base import BaseClassificationComponentTest


class ExtraTreesComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.96
    res["iris_n_calls"] = 9
    res["default_iris_iterative"] = res["default_iris"]
    res["default_iris_proba"] = 0.10053485167017469
    res["default_iris_sparse"] = 0.74
    res["default_digits"] = 0.9216757741347905
    res["digits_n_calls"] = 9
    res["default_digits_iterative"] = res["default_digits"]
    res["default_digits_iterative_places"] = 3
    res["default_digits_binary"] = 0.994535519125683
    res["default_digits_multilabel"] = 0.9983621593291405
    res["default_digits_multilabel_proba"] = 0.997710730679746

    sk_mod = sklearn.ensemble.ExtraTreesClassifier
    module = ExtraTreesClassifier
    step_hyperparameter = {
        "name": "n_estimators",
        "value": module.get_max_iter(),
    }



import sklearn.svm

from autosklearn.pipeline.components.regression.liblinear_svr import LibLinear_SVR

from .test_base import BaseRegressionComponentTest


class SupportVectorComponentTest(BaseRegressionComponentTest):
    __test__ = True

    res = dict()
    res["default_boston"] = 0.6768297818275556
    res["default_boston_places"] = 2
    res["default_boston_iterative"] = None
    res["default_boston_sparse"] = 0.12626519114138912
    res["default_boston_sparse_places"] = 2
    res["default_boston_iterative_sparse"] = None
    res["default_diabetes"] = 0.39152218711865661
    res["default_diabetes_iterative"] = None
    res["default_diabetes_sparse"] = 0.18704323088631891
    res["default_diabetes_iterative_sparse"] = None

    sk_mod = sklearn.svm.LinearSVR

    module = LibLinear_SVR



from typing import Dict, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class NoEncoding(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self, random_state: Optional[Union[int, np.random.RandomState]] = None
    ) -> None:
        pass

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "NoEncoding":
        self.preprocessor = "passthrough"
        self.fitted_ = True
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        return X

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "no encoding",
            "name": "No categorical variable encoding",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        return cs



import numpy as np
import scipy.sparse

from autosklearn.pipeline.components.data_preprocessing.minority_coalescense.minority_coalescer import (  # noqa: E501
    MinorityCoalescer,
)
from autosklearn.pipeline.components.data_preprocessing.minority_coalescense.no_coalescense import (  # noqa: E501
    NoCoalescence,
)

import unittest


class MinorityCoalescerTest(unittest.TestCase):
    def test_data_type_consistency(self):
        X = np.random.randint(3, 6, (3, 4))
        Y = MinorityCoalescer().fit_transform(X)
        self.assertFalse(scipy.sparse.issparse(Y))

        X = scipy.sparse.csc_matrix(
            ([3, 6, 4, 5], ([0, 1, 2, 1], [3, 2, 1, 0])), shape=(3, 4)
        )
        Y = MinorityCoalescer().fit_transform(X)
        self.assertTrue(scipy.sparse.issparse(Y))

    def test_no_coalescence(self):
        X = np.random.randint(0, 255, (3, 4))
        Y = NoCoalescence().fit_transform(X)
        np.testing.assert_array_almost_equal(Y, X)
        # Assert no copies were made
        self.assertEqual(id(X), id(Y))



import numpy as np
from scipy import sparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array


class CategoryShift(BaseEstimator, TransformerMixin):
    """Add 3 to every category."""

    def __init__(self, random_state=None):
        self.random_state = random_state

    def _convert_and_check_X(self, X):
        X_data = X.data if sparse.issparse(X) else X

        # Check if data is numeric and positive
        if X_data.dtype.kind not in set("buif") or np.nanmin(X_data) < 0:
            raise ValueError(
                "Categories should be non-negative numbers. "
                "NOTE: floats will be casted to integers."
            )

        # Use check_array to make sure we are using the right kind of sparse array
        # Notice that we cannot convert the array to integer right now. That would get
        # rid of the np.nans and we need them later on for the imputation.
        X = check_array(X, accept_sparse="csc", force_all_finite=False, copy=True)
        return X

    def fit(self, X, y=None):
        self._convert_and_check_X(X)
        return self

    def transform(self, X):
        X = self._convert_and_check_X(X)
        # Increment everything by three to account for the fact that
        # np.NaN will get an index of two, and coalesced values will get index of
        # one, index of zero is not assigned to also work with sparse data
        X_data = X.data if sparse.issparse(X) else X
        X_data += 3
        return X



import itertools

import numpy as np
from ConfigSpace.forbidden import ForbiddenAndConjunction, ForbiddenEqualsClause

from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    PREDICTIONS,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)


def get_match_array(pipeline, dataset_properties, include=None, exclude=None):
    sparse = dataset_properties.get("sparse")
    signed = dataset_properties.get("signed")

    # Duck typing, not sure if it's good...
    node_i_is_choice = []
    node_i_choices = []
    node_i_choices_names = []
    all_nodes = []
    for node_name, node in pipeline:
        all_nodes.append(node)
        is_choice = hasattr(node, "get_available_components")
        node_i_is_choice.append(is_choice)

        node_include = include.get(node_name) if include is not None else None
        node_exclude = exclude.get(node_name) if exclude is not None else None

        if is_choice:
            node_i_choices_names.append(
                list(
                    node.get_available_components(
                        dataset_properties, include=node_include, exclude=node_exclude
                    ).keys()
                )
            )
            node_i_choices.append(
                list(
                    node.get_available_components(
                        dataset_properties, include=node_include, exclude=node_exclude
                    ).values()
                )
            )

        else:
            node_i_choices.append([node])

    matches_dimensions = [len(choices) for choices in node_i_choices]
    # Start by allowing every combination of nodes. Go through all
    # combinations/pipelines and erase the illegal ones
    matches = np.ones(matches_dimensions, dtype=int)

    pipeline_idxs = [range(dim) for dim in matches_dimensions]
    for pipeline_instantiation_idxs in itertools.product(*pipeline_idxs):
        pipeline_instantiation = [
            node_i_choices[i][idx] for i, idx in enumerate(pipeline_instantiation_idxs)
        ]

        data_is_sparse = sparse
        dataset_is_signed = signed
        for node in pipeline_instantiation:
            node_input = node.get_properties()["input"]
            node_output = node.get_properties()["output"]

            # First check if these two instantiations of this node can work
            # together. Do this in multiple if statements to maintain
            # readability
            if (
                (data_is_sparse and SPARSE not in node_input)
                or not data_is_sparse
                and DENSE not in node_input
            ):
                matches[pipeline_instantiation_idxs] = 0
                break
            # No need to check if the node can handle SIGNED_DATA; this is
            # always assumed to be true
            elif not dataset_is_signed and UNSIGNED_DATA not in node_input:
                matches[pipeline_instantiation_idxs] = 0
                break

            if (
                (
                    INPUT in node_output
                    and DENSE not in node_output
                    and SPARSE not in node_output
                )
                or PREDICTIONS in node_output
                or (not data_is_sparse and DENSE in node_input and DENSE in node_output)
                or (data_is_sparse and SPARSE in node_input and SPARSE in node_output)
            ):
                # Don't change the data_is_sparse flag
                pass
            elif data_is_sparse and DENSE in node_output:
                data_is_sparse = False
            elif not data_is_sparse and SPARSE in node_output:
                data_is_sparse = True
            else:
                print(node)
                print("Data is sparse", data_is_sparse)
                print(node_input, node_output)
                raise ValueError("This combination is not allowed!")

            if PREDICTIONS in node_output:
                pass
            elif (
                INPUT in node_output
                and SIGNED_DATA not in node_output
                and UNSIGNED_DATA not in node_output
            ):
                pass
            elif SIGNED_DATA in node_output:
                dataset_is_signed = True
            elif UNSIGNED_DATA in node_output:
                dataset_is_signed = False
            else:
                print(node)
                print("Data is signed", dataset_is_signed)
                print(node_input, node_output)
                raise ValueError("This combination is not allowed!")

    return matches


def find_active_choices(
    matches, node, node_idx, dataset_properties, include=None, exclude=None
):
    if not hasattr(node, "get_available_components"):
        raise ValueError()
    available_components = node.get_available_components(
        dataset_properties, include=include, exclude=exclude
    )
    assert matches.shape[node_idx] == len(available_components), (
        matches.shape[node_idx],
        len(available_components),
    )

    choices = []
    for c_idx, component in enumerate(available_components):
        slices = tuple(
            slice(None) if idx != node_idx else slice(c_idx, c_idx + 1)
            for idx in range(len(matches.shape))
        )

        if np.sum(matches[slices]) > 0:
            choices.append(component)
    return choices


def add_forbidden(conf_space, pipeline, matches, dataset_properties, include, exclude):
    # Not sure if this works for 3D
    node_i_is_choice = []
    node_i_choices_names = []
    node_i_choices = []
    all_nodes = []
    for node_name, node in pipeline:
        all_nodes.append(node)
        is_choice = hasattr(node, "get_available_components")
        node_i_is_choice.append(is_choice)

        node_include = include.get(node_name) if include is not None else None
        node_exclude = exclude.get(node_name) if exclude is not None else None

        if is_choice:
            node_i_choices_names.append(
                node.get_available_components(
                    dataset_properties, include=node_include, exclude=node_exclude
                ).keys()
            )
            node_i_choices.append(
                node.get_available_components(
                    dataset_properties, include=node_include, exclude=node_exclude
                ).values()
            )

        else:
            node_i_choices_names.append([node_name])
            node_i_choices.append([node])

    # Find out all chains of choices. Only in such a chain its possible to
    # have several forbidden constraints
    choices_chains = []
    idx = 0
    while idx < len(pipeline):
        if node_i_is_choice[idx]:
            chain_start = idx
            idx += 1
            while idx < len(pipeline) and node_i_is_choice[idx]:
                idx += 1
            chain_stop = idx
            choices_chains.append((chain_start, chain_stop))
        idx += 1

    for choices_chain in choices_chains:
        constraints = set()

        chain_start = choices_chain[0]
        chain_stop = choices_chain[1]
        chain_length = chain_stop - chain_start

        # Add one to have also have chain_length in the range
        for sub_chain_length in range(2, chain_length + 1):
            for start_idx in range(chain_start, chain_stop - sub_chain_length + 1):
                indices = range(start_idx, start_idx + sub_chain_length)
                node_names = [pipeline[idx][0] for idx in indices]

                num_node_choices = []
                node_choice_names = []
                skip_array_shape = []

                for idx in indices:
                    node = all_nodes[idx]
                    available_components = node.get_available_components(
                        dataset_properties, include=node_i_choices_names[idx]
                    )
                    assert len(available_components) > 0, len(available_components)
                    skip_array_shape.append(len(available_components))
                    num_node_choices.append(range(len(available_components)))
                    node_choice_names.append([name for name in available_components])

                # Figure out which choices were already abandoned
                skip_array = np.zeros(skip_array_shape)
                for product in itertools.product(*num_node_choices):
                    for node_idx, choice_idx in enumerate(product):
                        node_idx += start_idx
                        slices_ = tuple(
                            slice(None)
                            if idx != node_idx
                            else slice(choice_idx, choice_idx + 1)
                            for idx in range(len(matches.shape))
                        )

                        if np.sum(matches[slices_]) == 0:
                            skip_array[product] = 1

                for product in itertools.product(*num_node_choices):
                    if skip_array[product]:
                        continue

                    slices = tuple(
                        slice(None)
                        if idx not in indices
                        else slice(
                            product[idx - start_idx], product[idx - start_idx] + 1
                        )
                        for idx in range(len(matches.shape))
                    )

                    # This prints the affected nodes
                    # print [node_choice_names[i][product[i]]
                    #        for i in range(len(product))], \
                    #     np.sum(matches[slices])

                    if np.sum(matches[slices]) == 0:
                        constraint = tuple(
                            [
                                (node_names[i], node_choice_names[i][product[i]])
                                for i in range(len(product))
                            ]
                        )

                        # Check if a more general constraint/forbidden clause
                        #  was already added
                        continue_ = False
                        for constraint_length in range(2, len(constraint)):
                            constr_starts = len(constraint) - constraint_length + 1
                            for constraint_start_idx in range(constr_starts):
                                constraint_end_idx = (
                                    constraint_start_idx + constraint_length
                                )
                                sub_constraint = constraint[
                                    constraint_start_idx:constraint_end_idx
                                ]
                                if sub_constraint in constraints:
                                    continue_ = True
                                    break
                            if continue_:
                                break
                        if continue_:
                            continue

                        constraints.add(constraint)

                        forbiddens = []
                        for i in range(len(product)):
                            forbiddens.append(
                                ForbiddenEqualsClause(
                                    conf_space.get_hyperparameter(
                                        node_names[i] + ":__choice__"
                                    ),
                                    node_choice_names[i][product[i]],
                                )
                            )
                        forbidden = ForbiddenAndConjunction(*forbiddens)
                        conf_space.add_forbidden_clause(forbidden)

    return conf_space



import numpy as np
from scipy import sparse

from autosklearn.pipeline.components.data_preprocessing.feature_type import (
    FeatTypeSplit,
)

import unittest


class PreprocessingPipelineTest(unittest.TestCase):
    def do_a_fit_transform(self, sparse_input):
        # X will be the input and Y is what we expect after transform. categ_feat stores
        # indicators of feature type (True if categorical, False if numerical)
        X, Y, categ_feat = [], [], []
        # Feature 1 (numerical):
        # This feature should be dropped due to lack of variance.
        categ_feat.append(False)
        X.append(np.array([3.14, 3.14, 3.14]).reshape(3, 1))
        Y.append(np.array([]).reshape(3, 0))
        # Feature 2 (numerical):
        # This feature should be normalized by having its mean subtracted from all
        # elements and by having them divided by the standard deviation.
        categ_feat.append(False)
        nf = np.array([1.0, 2.0, 3.0]).reshape(3, 1)  # mean = 2.
        sdev = np.sqrt(2.0 / 3.0)
        shift = (
            0 if sparse_input else 2.0
        )  # if sparse_input, there is no mean subtraction
        nft = (nf - shift) / sdev
        X.append(nf)
        Y.append(nft)
        # Feature 3 (numerical):
        # This feature has a missing value that should be imputed by the mean of the
        # other values (2.).
        # This feature should also be normalized as in the previous feature.
        categ_feat.append(False)
        X.append(np.array([1.0, np.nan, 3.0]).reshape(3, 1))
        Y.append(nft.copy())
        # Feature 4 (categorical)
        # This feature should be one hot encoded.
        categ_feat.append(True)
        X.append(np.array([1, 3, 2]).reshape(3, 1))
        Y.append(np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]]))
        # Feature 5 (categorical)
        # This feature should be one hot encoded. (A discontinuous category set or
        # a category 0 shouldn't be problems.)
        categ_feat.append(True)
        X.append(np.array([2, 1, 9]).reshape(3, 1))
        Y.append(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
        # Feature 6 (categorical)
        # This feature should be one hot encoded. The missing value gets imputed as
        # a category on its own.
        categ_feat.append(True)
        X.append(np.array([1, 1, np.nan]).reshape(3, 1))
        Y.append(np.array([[0, 1], [0, 1], [1, 0]]))
        # Combine datasets and shuffle columns:
        n_feats = len(categ_feat)
        random_order = np.random.choice(np.arange(n_feats), size=n_feats, replace=False)
        # Shuffle X according to random_order
        X = np.array(X)[random_order]
        X_comb = np.hstack(X)
        # Shuffle Y according to random_order and reorder it as the
        # PreprocessingPipeline does (i.e. categorical features come first in Y).

        categ_feat = {
            i: "categorical" if categ_feat[order] else "numerical"
            for i, order in enumerate(random_order)
        }
        cat_to_left_order = [
            index
            for col, index in sorted(
                [(col_type, i) for i, col_type in categ_feat.items()]
            )
        ]
        # Sort so that Y Matches the random ordering
        Y = [Y[n] for n in random_order]
        # Then move the categorical columns to the left
        Y = [Y[n] for n in cat_to_left_order]
        Y_comb = np.hstack(Y)
        # Data preprocessing
        DPP = FeatTypeSplit(feat_type=categ_feat)
        X_comb = sparse.csc_matrix(X_comb) if sparse_input else X_comb
        Y_comb_out_1 = DPP.fit_transform(X_comb)
        # Check if Y_comb_out is what we expect it to be:
        self.assertEqual(sparse_input, sparse.issparse(Y_comb_out_1))
        Y_comb_out_1 = Y_comb_out_1.todense() if sparse_input else Y_comb_out_1
        np.testing.assert_array_almost_equal(Y_comb_out_1, Y_comb)
        # Do it again, but using the already fitted pipeline
        Y_comb_out_2 = DPP.transform(X_comb)
        # Consistency check
        self.assertEqual(sparse_input, sparse.issparse(Y_comb_out_2))
        Y_comb_out_2 = Y_comb_out_2.todense() if sparse_input else Y_comb_out_2
        np.testing.assert_array_equal(Y_comb_out_1, Y_comb_out_2)

    def test_fit_transform(self):
        self.do_a_fit_transform(sparse_input=False)

    def test_fit_transform_sparse(self):
        self.do_a_fit_transform(sparse_input=True)

    def test_string_categories(self):
        # Numerical dataset (as used in NumericalPreprocessingPipelineTest)
        X_num = np.array(
            [
                [3.14, 1.0, 1.0],  # noqa : matrix legibility
                [3.14, 2.0, np.nan],  # noqa : matrix legibility
                [3.14, 3.0, 3.0],
            ]
        )  # noqa : matrix legibility
        # Categorical string dataset
        X_cat = np.array(
            [
                ["red", "medium", "small"],
                ["blue", "short", "big"],
                ["white", "tall", np.nan],
            ]
        )
        # Combined dataset with shuffled columns:
        X_comb = np.hstack((X_num, X_cat))
        categ_feat = [False] * 3 + [True] * 3
        random_order = np.random.choice(np.arange(6), size=6, replace=False)
        X_comb = X_comb[:, random_order]
        categ_feat = [categ_feat[order] for order in random_order]
        # Strings are not allowed, therefore:
        with self.assertRaises(ValueError):
            categ_feat = {
                i: "categorical" if feat else "numerical"
                for i, feat in enumerate(categ_feat)
            }
            FeatTypeSplit(feat_type=categ_feat).fit_transform(X_comb)



import autosklearn.pipeline.components.feature_preprocessing as fp

import unittest


class FeatureProcessingTest(unittest.TestCase):
    def test_get_available_components(self):
        # Target type
        for target_type, num_values in [("classification", 15), ("regression", 14)]:
            data_properties = {"target_type": target_type}

            available_components = fp.FeaturePreprocessorChoice(
                data_properties
            ).get_available_components(data_properties)

            self.assertEqual(len(available_components), num_values)

        # Multiclass
        data_properties = {"target_type": "classification", "multiclass": True}
        available_components = fp.FeaturePreprocessorChoice(
            data_properties
        ).get_available_components(data_properties)

        self.assertEqual(len(available_components), 15)

        # Multilabel
        data_properties = {"target_type": "classification", "multilabel": True}
        available_components = fp.FeaturePreprocessorChoice(
            data_properties
        ).get_available_components(data_properties)

        self.assertEqual(len(available_components), 12)



from typing import Dict, Optional, Tuple, Union

import itertools

import ConfigSpace.hyperparameters as CSH
import numpy as np
import pandas as pd
from ConfigSpace import EqualsCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from scipy.sparse import hstack
from sklearn.feature_extraction.text import TfidfVectorizer

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool


class TfidfEncoder(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        ngram_range_word: int = 1,
        ngram_range_char: int = 4,
        min_df: float = 0.0,
        max_df: float = 1.0,
        binary: bool = False,
        norm: str = "l2",
        sublinear_tf: bool = False,
        per_column: bool = False,
        analyzer: str = "char",
        random_state: Optional[Union[int, np.random.RandomState]] = None,
    ) -> None:
        self.random_state = random_state
        self.min_df = min_df
        self.max_df = max_df
        self.binary = binary
        self.norm = norm
        self.sublinear_tf = sublinear_tf
        self.per_column = per_column
        self.analyzer = analyzer
        self.ngram_range_word = ngram_range_word
        self.ngram_range_char = ngram_range_char

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "TfidfEncoder":
        if self.analyzer == "word":
            ngram_range = self.ngram_range_word
        elif self.analyzer == "char":
            ngram_range = self.ngram_range_char
        else:
            raise KeyError(f"Analyzer is not defined for {self.analyzer}")

        self.sublinear_tf = check_for_bool(self.sublinear_tf)
        self.binary = check_for_bool(self.binary)

        if isinstance(X, pd.DataFrame):
            X.fillna("", inplace=True)
            if self.per_column:
                self.preprocessor = {}

                for feature in X.columns:
                    vectorizer = TfidfVectorizer(
                        ngram_range=(ngram_range, ngram_range),
                        min_df=self.min_df,
                        max_df=self.max_df,
                        binary=self.binary,
                        norm=self.norm,
                        sublinear_tf=self.sublinear_tf,
                        analyzer=self.analyzer,
                    ).fit(X[feature])
                    self.preprocessor[feature] = vectorizer
            else:
                self.preprocessor = TfidfVectorizer(
                    ngram_range=(ngram_range, ngram_range),
                    min_df=self.min_df,
                    max_df=self.max_df,
                    binary=self.binary,
                    norm=self.norm,
                    sublinear_tf=self.sublinear_tf,
                    analyzer=self.analyzer,
                )
                all_text = itertools.chain.from_iterable(X[col] for col in X.columns)
                self.preprocessor = self.preprocessor.fit(all_text)

        else:
            raise ValueError(
                "Your text data is not encoded in a pandas.DataFrame\n"
                "Please make sure to use a pandas.DataFrame and ensure"
                "that the text features are encoded as strings."
            )
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        X.fillna("", inplace=True)
        if self.per_column:
            X_new = None
            if self.preprocessor is None:
                raise NotImplementedError()

            for feature in self.preprocessor:
                # the names in the dataframe must not change
                if X_new is None:
                    X_new = self.preprocessor[feature].transform(X[feature])
                else:
                    X_transformed = self.preprocessor[feature].transform(X[feature])
                    X_new = hstack([X_new, X_transformed])

            return X_new
        else:
            X_transformed = None
            if self.preprocessor is None:
                raise NotImplementedError()
            for feature in X.columns:
                if X_transformed is None:
                    X_transformed = self.preprocessor.transform(X[feature])
                else:
                    X_transformed += self.preprocessor.transform(X[feature])
            return X_transformed

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "TF/IDF",
            "name": "TF/IDF Encoder",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()

        hp_analyzer = CSH.CategoricalHyperparameter(
            name="analyzer",
            choices=["word", "char"],
            default_value="char",
        )

        hp_ngram_range_char = CSH.UniformIntegerHyperparameter(
            name="ngram_range_char",
            lower=2,
            upper=4,
            default_value=4,
        )

        hp_ngram_range_word = CSH.UniformIntegerHyperparameter(
            name="ngram_range_word",
            lower=1,
            upper=3,
            default_value=1,
        )

        hp_min_df = CSH.UniformFloatHyperparameter(
            name="min_df",
            lower=0.0,
            upper=0.3,
            default_value=0.0,
        )

        hp_max_df = CSH.UniformFloatHyperparameter(
            name="max_df", lower=0.7, upper=1.0, default_value=1.0
        )

        hp_binary = CSH.UnParametrizedHyperparameter(name="binary", value="False")

        hp_norm = CSH.CategoricalHyperparameter(
            name="norm", choices=["l2", "l1"], default_value="l2"
        )

        hp_sublinear_tf = CSH.UnParametrizedHyperparameter(
            name="sublinear_tf", value="False"
        )

        hp_per_column = CSH.CategoricalHyperparameter(
            name="per_column", choices=[True, False], default_value=False
        )

        cs.add_hyperparameters(
            [
                hp_analyzer,
                hp_ngram_range_char,
                hp_ngram_range_word,
                hp_max_df,
                hp_min_df,
                hp_binary,
                hp_norm,
                hp_sublinear_tf,
                hp_per_column,
            ]
        )
        cond_char = EqualsCondition(hp_ngram_range_char, hp_analyzer, "char")
        cond_word = EqualsCondition(hp_ngram_range_word, hp_analyzer, "word")
        cs.add_conditions([cond_char, cond_word])

        return cs



import re

import numpy as np
import pkg_resources

from autosklearn.util.dependencies import (
    IncorrectPackageVersionError,
    MissingPackageError,
    verify_packages,
)

import unittest
from unittest.mock import Mock, patch


@patch("pkg_resources.get_distribution")
class VerifyPackagesTests(unittest.TestCase):
    def test_existing_package(self, getDistributionMock):
        requirement = "package"
        distribution_mock = unittest.mock.Mock()
        getDistributionMock.return_value = distribution_mock
        distribution_mock.version = "1.0.0"

        verify_packages(requirement)

        getDistributionMock.assert_called_once_with("package")

    def test_missing_package(self, getDistributionMock):
        requirement = "package"

        getDistributionMock.side_effect = pkg_resources.DistributionNotFound()

        self.assertRaisesRegex(
            MissingPackageError,
            "Mandatory package 'package' not found",
            verify_packages,
            requirement,
        )

    @patch("importlib.import_module")
    def test_package_can_only_be_imported(self, import_mock, getDistributionMock):

        getDistributionMock.side_effect = pkg_resources.DistributionNotFound()
        package = unittest.mock.Mock()
        package.__version__ = np.__version__
        import_mock.return_value = package

        verify_packages("numpy")

    def test_correct_package_versions(self, getDistributionMock):
        requirement = "package==0.1.2\n" "package>0.1\n" "package>=0.1"

        moduleMock = Mock()
        moduleMock.version = "0.1.2"
        getDistributionMock.return_value = moduleMock

        verify_packages(requirement)

        getDistributionMock.assert_called_with("package")
        self.assertEqual(3, len(getDistributionMock.call_args_list))

    def test_wrong_package_version(self, getDistributionMock):
        requirement = "package>0.1.2"

        moduleMock = Mock()
        moduleMock.version = "0.1.2"
        getDistributionMock.return_value = moduleMock

        self.assertRaisesRegex(
            IncorrectPackageVersionError,
            re.escape(
                "found 'package' version 0.1.2 but requires package version >0.1.2"
            ),
            verify_packages,
            requirement,
        )

    def test_outdated_requirement(self, getDistributionMock):
        requirement = "package>=0.1"

        moduleMock = Mock()
        moduleMock.version = "0.0.9"
        getDistributionMock.return_value = moduleMock

        self.assertRaisesRegex(
            IncorrectPackageVersionError,
            re.escape(
                "found 'package' version 0.0.9 but requires package version >=0.1"
            ),
            verify_packages,
            requirement,
        )

    def test_too_fresh_requirement(self, getDistributionMock):
        requirement = "package==0.1.2"

        moduleMock = Mock()
        moduleMock.version = "0.1.3"
        getDistributionMock.return_value = moduleMock

        self.assertRaisesRegex(
            IncorrectPackageVersionError,
            re.escape(
                "found 'package' version 0.1.3 but requires package version ==0.1.2"
            ),
            verify_packages,
            requirement,
        )



import logging
import os

import autosklearn.metalearning.optimizers.metalearn_optimizer.metalearner as metalearner  # noqa: E501
import autosklearn.pipeline.classification
from autosklearn.metalearning.metalearning.meta_base import MetaBase

import unittest

logging.basicConfig()


class MetalearningConfiguration(unittest.TestCase):
    def test_metalearning_cs_size(self):
        self.cwd = os.getcwd()
        data_dir = os.path.dirname(__file__)
        data_dir = os.path.join(data_dir, "test_meta_base_data")
        os.chdir(data_dir)

        # Total: 176, categorical: 3, numerical: 7, string: 7
        total = 179
        num_numerical = 6
        num_string = 11
        num_categorical = 3
        for feat_type, cs_size in [
            ({"A": "numerical"}, total - num_string - num_categorical),
            ({"A": "categorical"}, total - num_string - num_numerical),
            ({"A": "string"}, total - num_categorical - num_numerical),
            ({"A": "numerical", "B": "categorical"}, total - num_string),
            ({"A": "numerical", "B": "string"}, total - num_categorical),
            ({"A": "categorical", "B": "string"}, total - num_numerical),
            ({"A": "categorical", "B": "string", "C": "numerical"}, total),
        ]:
            pipeline = autosklearn.pipeline.classification.SimpleClassificationPipeline(
                feat_type=feat_type
            )
            self.cs = pipeline.get_hyperparameter_search_space(feat_type=feat_type)

            self.logger = logging.getLogger()
            meta_base = MetaBase(self.cs, data_dir, logger=self.logger)
            self.meta_optimizer = metalearner.MetaLearningOptimizer(
                "233", self.cs, meta_base, logger=self.logger
            )
            self.assertEqual(
                len(self.meta_optimizer.configuration_space), cs_size, feat_type
            )



from typing import Optional

import resource
import sys

from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnRegressionAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool, check_none


class LibSVM_SVR(AutoSklearnRegressionAlgorithm):
    def __init__(
        self,
        kernel,
        C,
        epsilon,
        tol,
        shrinking,
        gamma=0.1,
        degree=3,
        coef0=0.0,
        verbose=False,
        max_iter=-1,
        random_state=None,
    ):
        self.kernel = kernel
        self.C = C
        self.epsilon = epsilon
        self.tol = tol
        self.shrinking = shrinking
        self.degree = degree
        self.gamma = gamma
        self.coef0 = coef0
        self.verbose = verbose
        self.max_iter = max_iter
        self.random_state = random_state
        self.estimator = None

    def fit(self, X, y):
        import sklearn.svm

        # Calculate the size of the kernel cache (in MB) for sklearn's LibSVM.
        # The cache size is calculated as 2/3 of the available memory
        # (which is calculated as the memory limit minus the used memory)
        try:
            # Retrieve memory limits imposed on the process
            soft, hard = resource.getrlimit(resource.RLIMIT_AS)

            if soft > 0:
                # Convert limit to units of megabytes
                soft /= 1024 * 1024

                # Retrieve memory used by this process
                maxrss = resource.getrusage(resource.RUSAGE_SELF)[2] / 1024

                # In MacOS, the MaxRSS output of resource.getrusage in bytes;
                # on other platforms, it's in kilobytes
                if sys.platform == "darwin":
                    maxrss = maxrss / 1024

                cache_size = (soft - maxrss) / 1.5

                if cache_size < 0:
                    cache_size = 200
            else:
                cache_size = 200
        except Exception:
            cache_size = 200

        self.C = float(self.C)
        self.epsilon = float(self.epsilon)
        self.tol = float(self.tol)
        self.shrinking = check_for_bool(self.shrinking)
        self.degree = int(self.degree)
        self.gamma = float(self.gamma)
        if check_none(self.coef0):
            self.coef0 = 0.0
        else:
            self.coef0 = float(self.coef0)
        self.verbose = int(self.verbose)
        self.max_iter = int(self.max_iter)

        self.estimator = sklearn.svm.SVR(
            kernel=self.kernel,
            C=self.C,
            epsilon=self.epsilon,
            tol=self.tol,
            shrinking=self.shrinking,
            degree=self.degree,
            gamma=self.gamma,
            coef0=self.coef0,
            cache_size=cache_size,
            verbose=self.verbose,
            max_iter=self.max_iter,
        )
        self.scaler = sklearn.preprocessing.StandardScaler(copy=True)

        # Convert y to be at least 2d for the scaler
        # [1,1,1] -> [[1], [1], [1]]
        if y.ndim == 1:
            y = y.reshape((-1, 1))

        y_scaled = self.scaler.fit_transform(y)

        # Flatten: [[0], [0], [0]] -> [0, 0, 0]
        if y_scaled.ndim == 2 and y_scaled.shape[1] == 1:
            y_scaled = y_scaled.flatten()

        self.estimator.fit(X, y_scaled)

        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        if self.scaler is None:
            raise NotImplementedError
        y_pred = self.estimator.predict(X)

        inverse = self.scaler.inverse_transform(y_pred)

        # Flatten: [[0], [0], [0]] -> [0, 0, 0]
        if inverse.ndim == 2 and inverse.shape[1] == 1:
            inverse = inverse.flatten()

        return inverse

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "SVR",
            "name": "Support Vector Regression",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "prefers_data_normalized": True,
            "is_deterministic": True,
            "input": (SPARSE, DENSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        C = UniformFloatHyperparameter(
            name="C", lower=0.03125, upper=32768, log=True, default_value=1.0
        )
        # Random Guess
        epsilon = UniformFloatHyperparameter(
            name="epsilon", lower=0.001, upper=1, default_value=0.1, log=True
        )

        kernel = CategoricalHyperparameter(
            name="kernel",
            choices=["linear", "poly", "rbf", "sigmoid"],
            default_value="rbf",
        )
        degree = UniformIntegerHyperparameter(
            name="degree", lower=2, upper=5, default_value=3
        )

        gamma = UniformFloatHyperparameter(
            name="gamma", lower=3.0517578125e-05, upper=8, log=True, default_value=0.1
        )

        # TODO this is totally ad-hoc
        coef0 = UniformFloatHyperparameter(
            name="coef0", lower=-1, upper=1, default_value=0
        )
        # probability is no hyperparameter, but an argument to the SVM algo
        shrinking = CategoricalHyperparameter(
            name="shrinking", choices=["True", "False"], default_value="True"
        )
        tol = UniformFloatHyperparameter(
            name="tol", lower=1e-5, upper=1e-1, default_value=1e-3, log=True
        )
        max_iter = UnParametrizedHyperparameter("max_iter", -1)

        cs = ConfigurationSpace()
        cs.add_hyperparameters(
            [C, kernel, degree, gamma, coef0, shrinking, tol, max_iter, epsilon]
        )

        degree_depends_on_poly = EqualsCondition(degree, kernel, "poly")
        gamma_depends_on_kernel = InCondition(
            child=gamma, parent=kernel, values=("poly", "rbf")
        )
        coef0_depends_on_kernel = InCondition(
            child=coef0, parent=kernel, values=("poly", "sigmoid")
        )
        cs.add_conditions(
            [degree_depends_on_poly, gamma_depends_on_kernel, coef0_depends_on_kernel]
        )

        return cs



import sklearn.ensemble

from autosklearn.pipeline.components.classification.adaboost import AdaboostClassifier

from .test_base import BaseClassificationComponentTest


class AdaBoostComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.93999999999999995
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 0.22452300738472031
    res["default_iris_sparse"] = 0.85999999999999999
    res["default_digits"] = 0.6879174256223437
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.98299939283545845
    res["default_digits_multilabel"] = -1
    res["default_digits_multilabel_proba"] = -1

    sk_mod = sklearn.ensemble.AdaBoostClassifier

    module = AdaboostClassifier



from typing import Type

import os
from collections import OrderedDict

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter

from autosklearn.askl_typing import FEAT_TYPE_TYPE

from ..base import (
    AutoSklearnChoice,
    AutoSklearnRegressionAlgorithm,
    ThirdPartyComponents,
    _addons,
    find_components,
)

regressor_directory = os.path.split(__file__)[0]
_regressors = find_components(
    __package__, regressor_directory, AutoSklearnRegressionAlgorithm
)
additional_components = ThirdPartyComponents(AutoSklearnRegressionAlgorithm)
_addons["regression"] = additional_components


def add_regressor(regressor: Type[AutoSklearnRegressionAlgorithm]) -> None:
    additional_components.add_component(regressor)


class RegressorChoice(AutoSklearnChoice):
    @classmethod
    def get_components(cls):
        components = OrderedDict()
        components.update(_regressors)
        components.update(additional_components.components)
        return components

    @classmethod
    def get_available_components(
        cls, dataset_properties=None, include=None, exclude=None
    ):
        available_comp = cls.get_components()
        components_dict = OrderedDict()
        if dataset_properties is None:
            dataset_properties = {}

        if include is not None and exclude is not None:
            raise ValueError(
                "The argument include and exclude cannot be used together."
            )

        if include is not None:
            for incl in include:
                if incl not in available_comp:
                    raise ValueError(
                        "Trying to include unknown component: " "%s" % incl
                    )

        for name in available_comp:
            if include is not None and name not in include:
                continue
            elif exclude is not None and name in exclude:
                continue

            entry = available_comp[name]

            # Avoid infinite loop
            if entry == RegressorChoice:
                continue

            if entry.get_properties()["handles_regression"] is False:
                continue
            if (
                dataset_properties.get("multioutput") is True
                and entry.get_properties()["handles_multioutput"] is False
            ):
                continue
            components_dict[name] = entry

        return components_dict

    def get_hyperparameter_search_space(
        self,
        feat_type: FEAT_TYPE_TYPE,
        dataset_properties=None,
        default=None,
        include=None,
        exclude=None,
    ):
        if include is not None and exclude is not None:
            raise ValueError(
                "The argument include and exclude cannot be used together."
            )

        cs = ConfigurationSpace()

        # Compile a list of all estimator objects for this problem
        available_estimators = self.get_available_components(
            dataset_properties=dataset_properties, include=include, exclude=exclude
        )

        if len(available_estimators) == 0:
            raise ValueError("No regressors found")

        if default is None:
            defaults = ["random_forest", "support_vector_regression"] + list(
                available_estimators.keys()
            )
            for default_ in defaults:
                if default_ in available_estimators:
                    if include is not None and default_ not in include:
                        continue
                    if exclude is not None and default_ in exclude:
                        continue
                    default = default_
                    break

        estimator = CategoricalHyperparameter(
            "__choice__", list(available_estimators.keys()), default_value=default
        )
        cs.add_hyperparameter(estimator)
        for estimator_name in available_estimators.keys():
            estimator_configuration_space = available_estimators[
                estimator_name
            ].get_hyperparameter_search_space(
                feat_type=feat_type, dataset_properties=dataset_properties
            )
            parent_hyperparameter = {"parent": estimator, "value": estimator_name}
            cs.add_configuration_space(
                estimator_name,
                estimator_configuration_space,
                parent_hyperparameter=parent_hyperparameter,
            )

        return cs

    def estimator_supports_iterative_fit(self):
        return hasattr(self.choice, "iterative_fit")

    def get_max_iter(self):
        if self.estimator_supports_iterative_fit():
            return self.choice.get_max_iter()
        else:
            raise NotImplementedError()

    def get_current_iter(self):
        if self.estimator_supports_iterative_fit():
            return self.choice.get_current_iter()
        else:
            raise NotImplementedError()

    def iterative_fit(self, X, y, n_iter=1, **fit_params):
        # Allows to use check_is_fitted on the choice object
        self.fitted_ = True
        if fit_params is None:
            fit_params = {}
        return self.choice.iterative_fit(X, y, n_iter=n_iter, **fit_params)

    def configuration_fully_fitted(self):
        return self.choice.configuration_fully_fitted()
